#! /usr/bin/env python

"""This module contains code to locate the appropriate PSF library files
to use for a given simulation. It supports the selection of one PSF
"core" file, which is assumed to contain a 3D array of PSFs that is read
into a ``griddedPSFmodel`` instance, as well as a single PSF "wings" file,
which contains a single PSF instance.

For both of these files, once they are identified, they are read in via
the appropriate mechanisms for the data they contain, and the resulting
object is returned.

Author
------

    - Bryan Hilbert

Use
---

    This module can be imported and called as such:
    ::
        from mirage.psf import psf_selection
        library = psf_selection.get_gridded_psf_library('nircam', 'nrcb1',
                                                        'f200w', 'clear',
                                                        'predicted', 0,
                                                        '/path/to/library/')
"""


from copy import copy
from glob import glob
import os
import warnings

from astropy.io import fits
import numpy as np
from webbpsf.utils import to_griddedpsfmodel

from mirage.utils.constants import NIRISS_PUPIL_WHEEL_FILTERS, NIRCAM_PUPIL_WHEEL_FILTERS
from mirage.utils.utils import expand_environment_variable



def confirm_gridded_properties(filename, instrument, detector, filtername, pupilname,
                               wavefront_error_type, wavefront_error_group, file_path,
                               extname='PRIMARY'):
    """Examine the header of the gridded PSF model file to confirm that
    the properties of the data match those expected.

    Parameters
    ----------
    filename : str
        Base name of the PSF library file to be checked

    instrument : str
        Name of instrument the PSFs are from

    detector : str
        Name of the detector within ```instrument```

    filtername : str
        Name of filter used for PSF library creation

    pupilname : str
        Name of pupil wheel element used for PSF library creation

    wavefront_error_type : str
        Wavefront error. Can be 'predicted' or 'requirements'

    wavefront_error_group : int
        Wavefront error realization group. Must be an integer from 0 - 9.

    file_path : str
        Path pointing to the location of the PSF library

    extname : str
        Name of the extension within ``filename`` to check

    Returns
    -------
    full_filename : str
        Full path and filename if the file properties are as expected.
        None if the properties do not match.
    """

    # Determine if the PSF path is default or not
    mirage_dir = expand_environment_variable('MIRAGE_DATA')
    default_psf = file_path == os.path.join(mirage_dir,
                                               '{}/gridded_psf_library'.format(instrument.lower()))

    full_filename = os.path.join(file_path, filename)
    with fits.open(full_filename) as hdulist:
        header = hdulist[extname.upper()].header

    inst = header['INSTRUME']
    try:
        det = header['DETECTOR']
    except KeyError:
        det = header['DET_NAME']
    filt = header['FILTER']
    try:
        pupil = header['PUPIL']
    except KeyError:
        # If no pupil mask value is present, then assume the CLEAR is
        # being used
        if instrument.upper() == 'NIRCAM':
            pupil = 'CLEAR'
        elif instrument.upper() == 'NIRISS':
            pupil = 'CLEARP'

    # NIRISS has many filters in the pupil wheel. WebbPSF does
    # not make a distinction, but Mirage does. Adjust the info
    # to match Mirage's expectations
    if inst.upper() == 'NIRISS' and filt in NIRISS_PUPIL_WHEEL_FILTERS:
        save_filt = copy(filt)
        if pupil == 'CLEARP':
            filt = 'CLEAR'
        else:
            raise ValueError(('Pupil value is something other than '
                              'CLEARP, but the filter being used is '
                              'in the pupil wheel.'))
        pupil = save_filt

    # Same for NIRCam
    if inst.upper() == 'NIRCAM' and filt in NIRCAM_PUPIL_WHEEL_FILTERS:
        save_filt = copy(filt)
        if pupil == 'CLEAR':
            if save_filt[0:2] == 'F4':
                filt = 'F444W'
            elif save_filt[0:2] == 'F3':
                filt = 'F322W2'
            elif save_filt[0:2] == 'F1':
               filt = 'F150W2'
        else:
            raise ValueError(('Pupil value is something other than '
                              'CLEAR, but the filter being used is '
                              'in the pupil wheel.'))
        pupil = save_filt

    opd_file = header['OPD_FILE']
    if default_psf:
        if 'predicted' in opd_file:
            wfe_type = 'predicted'
        elif 'requirements' in opd_file:
            wfe_type = 'requirements'
        realization = header['OPDSLICE']

    # make the check below pass for FGS
    if instrument.lower() == 'fgs':
        pupil = 'N/A'
        pupilname = 'N/A'
        filt = 'N/A'
        filtername = 'N/A'

    match = inst.lower() == instrument.lower() and \
            det.lower() == detector.lower() and \
            filt.lower() == filtername.lower() and \
            pupil.lower() == pupilname.lower()
    if match and not default_psf:
        return full_filename
    elif match and wfe_type == wavefront_error_type.lower() and \
        realization == wavefront_error_group:
            return full_filename
    else:
        return None


def get_gridded_psf_library(instrument, detector, filtername, pupilname, wavefront_error,
                            wavefront_error_group, library_path):
    """Find the filename for the appropriate gridded PSF library and
    read it in to a griddedPSFModel

    Parameters
    ----------
    instrument : str
        Name of instrument the PSFs are from

    detector : str
        Name of the detector within ```instrument```

    filtername : str
        Name of filter used for PSF library creation

    pupilname : str
        Name of pupil wheel element used for PSF library creation

    wavefront_error : str
        Wavefront error. Can be 'predicted' or 'requirements'

    wavefront_error_group : int
        Wavefront error realization group. Must be an integer from 0 - 9.

    library_path : str
        Path pointing to the location of the PSF library

    Returns:
    --------
    library : photutils.griddedPSFModel
        Object containing PSF library

    """
    # First, as a way to save time, let's assume a file naming convention
    # and search for the appropriate file that way. If we find a match,
    # confirm the properties of the file via the header. This way we don't
    # need to open and examine every file in the gridded library, which
    # saves at least a handful of seconds.
    if instrument.lower() == 'fgs':
        default_file_pattern = '{}_{}_fovp*_samp*_npsf*_{}_realization{}.fits'.format(instrument.lower(),
                                                                                        detector.lower(),
                                                                                        wavefront_error.lower(),
                                                                                        wavefront_error_group)
    else:
        default_file_pattern = '{}_{}_{}_{}_fovp*_samp*_npsf*_{}_realization{}.fits'.format(instrument.lower(),
                                                                                        detector.lower(),
                                                                                        filtername.lower(),
                                                                                        pupilname.lower(),
                                                                                        wavefront_error.lower(),
                                                                                        wavefront_error_group)
    default_matches = glob(os.path.join(library_path, default_file_pattern))

    library_file = None
    if len(default_matches) == 1:
        library_file = confirm_gridded_properties(default_matches[0], instrument, detector, filtername,
                                                  pupilname, wavefront_error, wavefront_error_group,
                                                  library_path)

    # If the above search found no matching files, or multiple matching
    # files (based only on filename), or if the matching file's gridded
    # PSF model properties don't match what's expected, then resort to
    # opening and examining all files in the library.
    if library_file is None:
        library_file = get_library_file(instrument, detector, filtername, pupilname,
                                        wavefront_error, wavefront_error_group, library_path)

    print("PSFs will be generated using: {}".format(os.path.abspath(library_file)))

    try:
        library = to_griddedpsfmodel(library_file)
    except KeyError:
        # Handle input ITM images
        itm_sim = fits.getval(library_file, 'ORIGIN')
        if itm_sim:
            library = _load_itm_library(library_file)

    except OSError:
        print("OSError: Unable to open {}.".format(library_file))
    return library


def get_library_file(instrument, detector, filt, pupil, wfe, wfe_group,
                     library_path, wings=False, segment_id=None):
    """Given an instrument and filter name along with the path of
    the PSF library, find the appropriate library file to load.

    Parameters
    -----------
    instrument : str
        Name of instrument the PSFs are from

    detector : str
        Name of the detector within ```instrument```

    filt : str
        Name of filter used for PSF library creation

    pupil : str
        Name of pupil wheel element used for PSF library creation

    wfe : str
        Wavefront error. Can be 'predicted' or 'requirements'

    wfe_group : int
        Wavefront error realization group. Must be an integer from 0 - 9.

    library_path : str
        Path pointing to the location of the PSF library

    wings : bool, optional
        Must the library file contain PSF wings or PSF cores? Default is False.

    segment_id : int or None, optional
        If specified, returns a segment PSF library file and denotes the ID
        of the mirror segment

    Returns
    --------
    matches : str
        Name of the PSF library file for the instrument and filter name
    """
    psf_files = glob(os.path.join(library_path, '*.fits'))

    # Determine if the PSF path is default or not
    mirage_dir = expand_environment_variable('MIRAGE_DATA')
    gridded_dir = os.path.join(mirage_dir, '{}/gridded_psf_library'.format(instrument.lower()))
    if wings:
        gridded_dir = os.path.join(gridded_dir, 'psf_wings')
    default_psf = library_path == gridded_dir

    # Create a dictionary of header information for all PSF library files
    matches = []

    instrument = instrument.upper()
    detector = detector.upper()
    filt = filt.upper()
    pupil = pupil.upper()
    wfe = wfe.lower()

    # set default
    file_wfe = ''

    # handle the NIRISS NRM case
    if pupil == 'NRM':
        pupil = 'MASK_NRM'

    for filename in psf_files:
        try:
            header = fits.getheader(filename)

            # Determine if it is an ITM file
            itm_sim = header.get('ORIGIN', '') == 'ITM'

            # Compare the header entries to the user input
            file_inst = header['INSTRUME'].upper()
            try:
                file_det = header['DETECTOR'].upper()
            except KeyError:
                file_det = header['DET_NAME'].upper()
            file_filt = header['FILTER'].upper()

            try:
                file_pupil = header['PUPIL'].upper()
            except KeyError:
                # If no pupil mask value is present, then assume the CLEAR is
                # being used
                if file_inst.upper() == 'NIRCAM':
                    file_pupil = 'CLEAR'
                elif file_inst.upper() == 'NIRISS':
                    try:
                        file_pupil = header['PUPIL'].upper()  # can be 'MASK_NRM'
                    except KeyError:
                        file_pupil = 'CLEARP'

            # NIRISS has many filters in the pupil wheel. WebbPSF does
            # not make a distinction, but Mirage does. Adjust the info
            # to match Mirage's expectations
            if file_inst.upper() == 'NIRISS' and file_filt in NIRISS_PUPIL_WHEEL_FILTERS:
                save_filt = copy(file_filt)
                if file_pupil == 'CLEARP':
                    file_filt = 'CLEAR'
                else:
                    raise ValueError(('Pupil value is something other than '
                                      'CLEARP, but the filter being used is '
                                      'in the pupil wheel.'))
                file_pupil = save_filt

            # Same for NIRCam
            if file_inst.upper() == 'NIRCAM' and file_filt in NIRCAM_PUPIL_WHEEL_FILTERS:
                save_filt = copy(file_filt)
                if file_pupil == 'CLEAR':
                    if save_filt[0:2] == 'F4':
                        file_filt = 'F444W'
                    elif save_filt[0:2] == 'F3':
                        file_filt = 'F322W2'
                    elif save_filt[0:2] == 'F1':
                        file_filt = 'F150W2'
                else:
                    raise ValueError(('Pupil value is something other than '
                                      'CLEAR, but the filter being used is '
                                      'in the pupil wheel.'))
                file_pupil = save_filt

            if segment_id is None and not itm_sim:
                opd = header['OPD_FILE']
                if 'requirements' in opd:
                    file_wfe = 'requirements'
                elif 'predicted' in opd:
                    file_wfe = 'predicted'

                file_wfe_grp = header['OPDSLICE']

            if segment_id is not None:
                segment_id = int(segment_id)
                file_segment_id = int(header['SEGID'])

            # allow check below to pass for FGS
            if instrument.lower() == 'fgs':
                file_filt = 'N/A'
                filt = 'N/A'
                file_pupil = 'N/A'
                pupil = 'N/A'

            # Evaluate if the file matches the given parameters
            match = (file_inst == instrument
                     and file_det == detector
                     and file_filt == filt
                     and file_pupil == pupil
                     and file_wfe == wfe)

            if not wings and segment_id is None and not itm_sim and default_psf:
                match = match and file_wfe_grp == wfe_group
            if segment_id is not None:
                match = match and file_segment_id == segment_id
            elif not itm_sim and default_psf:
                match = match and file_wfe == wfe

            # If so, add to the list of all matches
            if match:
                matches.append(filename)
        except KeyError as e:
            warnings.warn('While searching for PSF file, error raised when examining {}:\n{}\nContinuing.'.format(os.path.basename(filename), e))
            continue

    # Find files matching the requested inputs
    if len(matches) == 1:
        return matches[0]
    elif len(matches) == 0:
        print('Requested parameters:\ninstrument {}\ndetector {}\nfilt {}\npupil {}\nwfe {}\n'
              'wfe_group {}\nlibrary_path {}\n'.format(instrument, detector, filt, pupil, wfe,
                                                       wfe_group, library_path))
        raise ValueError("No PSF library file found matching requested parameters.")
    elif len(matches) > 1:
        raise ValueError("More than one PSF library file matches requested parameters: {}".format(matches))


def get_psf_wings(instrument, detector, filtername, pupilname, wavefront_error, wavefront_error_group,
                  library_path):
    """Locate the file containing PSF wing image and read them in. The
    idea is that there will only be one file for a given detector/filter/
    pupil/WFE/realization combination. This file will contain a PSF
    sampled at detector resolution and covering some large area in pixels.
    Later, when making the seed image, the appropriate subarray will be
    pulled out of this array for each input source depending on its
    magnitude.

    Parameters
    ----------
    instrument : str
        Name of instrument the PSFs are from

    detector : str
        Name of the detector within ```instrument```

    filtername : str
        Name of filter used for PSF library creation

    pupilname : str
        Name of pupil wheel element used for PSF library creation

    wavefront_error : str
        Wavefront error. Can be 'predicted' or 'requirements'

    wavefront_error_group : int
        Wavefront error realization group. Must be an integer from 0 - 9.

    library_path : str
        Path pointing to the location of the PSF library

    Returns
    -------
    psf_wings : numpy.ndarray
        Array containing the PSF wing data. Note that the outermost row
        and column are not returned, in order to avoid edge effects

    """
    # First, as a way to save time, let's assume a file naming convention
    # and search for the appropriate file that way. If we find a match,
    # confirm the properties of the file via the header. This way we don't
    # need to open and examine every file in the gridded library, which
    # saves at least a handful of seconds.
    default_file_pattern = '{}_{}_{}_{}_fovp*_samp*_{}_realization{}.fits'.format(instrument.lower(),
                                                                                  detector.lower(),
                                                                                  filtername.lower(),
                                                                                  pupilname.lower(),
                                                                                  wavefront_error.lower(),
                                                                                  wavefront_error_group)
    default_matches = glob(os.path.join(library_path, default_file_pattern))

    wings_file = None
    if len(default_matches) == 1:
        wings_file = confirm_gridded_properties(default_matches[0], instrument, detector, filtername,
                                                pupilname, wavefront_error, wavefront_error_group,
                                                library_path, extname='DET_DIST')

    # If the above search found no matching files, or multiple matching
    # files (based only on filename), or if the matching file's gridded
    # PSF model properties don't match what's expected, then resort to
    # opening and examining all files in the library.
    if wings_file is None:
        # Find the file containing the PSF wings
        wings_file = get_library_file(instrument, detector, filtername, pupilname,
                                      wavefront_error, wavefront_error_group, library_path, wings=True)

    print("PSF wings will be from: {}".format(os.path.basename(wings_file)))
    with fits.open(wings_file) as hdulist:
        psf_wing = hdulist['DET_DIST'].data
    # Crop the outer row and column in order to remove any potential edge
    # effects leftover from creation
    psf_wing = psf_wing[1:-1, 1:-1]

    for shape in psf_wing.shape:
        if shape % 2 == 0:
            print(("WARNING: PSF wing file contains an even number of rows or columns. "
                   "These must be even."))
            raise ValueError
    return psf_wing


def _load_itm_library(library_file):
    """Load ITM FITS file

    Parameters
    ----------
    library_path : str
        Path pointing to the location of the PSF library

    Returns
    -------
    library : photutils.griddedPSFModel
        Object containing PSF library
    """
    data = fits.getdata(library_file)
    hdr = fits.getheader(library_file)
    if data.shape == (2048, 2048):
        # TODO: Remove when Shannon adds her 3rd dimension check
        # Add (empty) 3rd dimension to the data
        data = np.array([data])

        # Add PSF location and oversampling keywords
        hdr['DET_YX0'] = ('(1023, 1023)', "The #0 PSF's (y,x) detector pixel position")
        hdr['OVERSAMP'] = (1, 'Oversampling factor for FFTs in computation')

        # Convert to HDUList and create library
        phdu = fits.PrimaryHDU(data, hdr)
        hdulist = fits.HDUList(phdu)
        library = to_griddedpsfmodel(hdulist)

        return library
    else:
        raise ValueError('Expecting ITM data of size (2048, 2048), not {}'.format(data.shape))



#! /usr/bin/env python

'''
Randomly change the magnitudes in input catalogs, to
help seed image testing and development
'''

from astropy.io import ascii
from numpy.random import normal

#infile = 'seed_im_from_catalog_test_ptsrc_catalog.list'
#infile = 'seed_im_from_catalog_test_galaxies_catalog.list'
#outfile = infile[0:-5] + '_filter2.list'
infile = 'seed_im_from_catalog_test_ptsrc_bright_catalog.list'
outfile = infile

cat = ascii.read(infile)

mags = cat['magnitude'].data
add = mags - 2.
#add = normal(mags,0.5)
cat['magnitude'] = add

ascii.write(cat,outfile,overwrite=True)



#! /usr/bin/env python

"""
Utility for saving seed images
"""
from astropy.io import fits
import numpy as np

import mirage


def save(seed_image, param_file, parameters, photflam, photfnu, pivot_wavelength,
         fullframe_size, nominal_dimensions, coord_adjust, grism_direct_factor,
         filename=None, segmentation_map=None, frametime=None, base_unit='ADU'):
    """Save a seed image
    """
    arrayshape = seed_image.shape
    if len(arrayshape) == 2:
        units = '{}/sec'.format(base_unit)
        yd, xd = arrayshape
        tgroup = 0.
        print('Seed image is 2D.')
    elif len(arrayshape) == 3:
        units = base_unit
        g, yd, xd = arrayshape
        tgroup = frametime * (parameters['Readout']['nframe'] + parameters['Readout']['nskip'])
        print('Seed image is 3D.')
    elif len(arrayshape) == 4:
        units = base_unit
        integ, g, yd, xd = arrayshape
        tgroup = frametime * (parameters['Readout']['nframe'] + parameters['Readout']['nskip'])
        print('Seed image is 4D.')

    xcent_fov = xd / 2
    ycent_fov = yd / 2

    kw = {}
    kw['xcenter'] = xcent_fov
    kw['ycenter'] = ycent_fov
    kw['units'] = units
    kw['TGROUP'] = tgroup
    if parameters['Readout']['pupil'][0].upper() == 'F':
        usefilt = 'pupil'
    else:
        usefilt = 'filter'

    if filename is None:
        basename = os.path.join(parameters['Output']['directory'],
                                parameters['Output']['file'][0:-5].split('/')[-1])
        filename = '{}_{}_seed_image.fits'.format(basename, parameters['Readout'][usefilt])

    # Set FGS filter to "N/A" in the output file
    # as this is the value DMS looks for.
    if parameters['Readout'][usefilt] == "NA":
        parameters['Readout'][usefilt] = "N/A"
    kw['filter'] = parameters['Readout'][usefilt]
    kw['PHOTFLAM'] = photflam
    kw['PHOTFNU'] = photfnu
    kw['PHOTPLAM'] = pivot_wavelength * 1.e4  # put into angstroms
    kw['NOMXDIM'] = nominal_dimensions[1]
    kw['NOMYDIM'] = nominal_dimensions[0]
    kw['NOMXSTRT'] = np.int(coord_adjust['xoffset'] + 1)
    kw['NOMXEND'] = np.int(nominal_dimensions[1] + coord_adjust['xoffset'])
    kw['NOMYSTRT'] = np.int(coord_adjust['yoffset'] + 1)
    kw['NOMYEND'] = np.int(nominal_dimensions[0] + coord_adjust['yoffset'])

    # Files/inputs used during seed image production
    kw['YAMLFILE'] = param_file
    kw['GAINFILE'] = parameters['Reffiles']['gain']
    kw['DISTORTN'] = parameters['Reffiles']['astrometric']
    kw['IPC'] = parameters['Reffiles']['ipc']
    kw['PIXARMAP'] = parameters['Reffiles']['pixelAreaMap']
    kw['CROSSTLK'] = parameters['Reffiles']['crosstalk']
    kw['FLUX_CAL'] = parameters['Reffiles']['flux_cal']
    kw['FTHRUPUT'] = parameters['Reffiles']['filter_throughput']
    kw['PTSRCCAT'] = parameters['simSignals']['pointsource']
    kw['GALAXCAT'] = parameters['simSignals']['galaxyListFile']
    kw['EXTNDCAT'] = parameters['simSignals']['extended']
    kw['MTPTSCAT'] = parameters['simSignals']['movingTargetList']
    kw['MTSERSIC'] = parameters['simSignals']['movingTargetSersic']
    kw['MTEXTEND'] = parameters['simSignals']['movingTargetExtended']
    kw['NONSDRAL'] = parameters['simSignals']['movingTargetToTrack']
    kw['BKGDRATE'] = parameters['simSignals']['bkgdrate']
    kw['TRACKING'] = parameters['Telescope']['tracking']
    kw['POISSON'] = parameters['simSignals']['poissonseed']
    kw['PSFWFE'] = parameters['simSignals']['psfwfe']
    kw['PSFWFGRP'] = parameters['simSignals']['psfwfegroup']
    kw['MRGEVRSN'] = mirage.__version__

    # Seed images provided to disperser are always embedded in an array
    # with dimensions equal to full frame * self.grism_direct_factor
    if parameters['Inst']['mode'] in ['wfss', 'ts_wfss']:
        kw['NOMXDIM'] = fullframe_size
        kw['NOMYDIM'] = fullframe_size
        kw['NOMXSTRT'] = np.int(fullframe_size * (grism_direct_factor - 1) / 2.)
        kw['NOMXEND'] = kw['NOMXSTRT'] + fullframe_size - 1
        kw['NOMYSTRT'] = np.int(fullframe_size * (grism_direct_factor - 1) / 2.)
        kw['NOMYEND'] = kw['NOMYSTRT'] + fullframe_size - 1

    kw['GRISMPAD'] = grism_direct_factor
    seedinfo = kw
    save_single_fits(seed_image, filename, key_dict=kw, image2=segmentation_map, image2type='SEGMAP')

    # Keep this print statement in the code that calls this function
    #print("Seed image and segmentation map saved as {}".format(self.seed_file))
    #print("Seed image, segmentation map, and metadata available as:")
    #print("self.seedimage, self.seed_segmap, self.seedinfo.")
    return filename, seedinfo


def save_single_fits(image, name, key_dict=None, image2=None, image2type=None):
        # Save an array into the first extension of a fits file
        h0 = fits.PrimaryHDU()
        h1 = fits.ImageHDU(image, name='DATA')
        if image2 is not None:
            h2 = fits.ImageHDU(image2)
            if image2type is not None:
                h2.header['EXTNAME'] = image2type

        # if a keyword dictionary is provided, put the
        # keywords into the 0th and 1st extension headers
        if key_dict is not None:
            for key in key_dict:
                h0.header[key] = key_dict[key]
                h1.header[key] = key_dict[key]

        if image2 is None:
            hdulist = fits.HDUList([h0, h1])
        else:
            hdulist = fits.HDUList([h0, h1, h2])
        hdulist.writeto(name, overwrite=True)




"""Generate random deployment errors  and apply them to an OPD.

Authors
-------
    - Lauren Chambers

Use
---
    ::
        deployment_errors = get_deployment_errors(out_dir=out_dir)
        deployment_errors = reduce_deployment_errors(deployment_errors, reduction_factor=0.2, out_dir=out_dir)
        ote, segment_tilts = apply_deployment_errors(ote, deployment_errors, out_dir=out_dir)
        ote = remove_piston_tip_tilt(ote, out_dir=out_dir)
"""

import os
import time
import yaml

from astropy.io import fits
import numpy as np
import webbpsf


def load_ote_from_deployment_yaml(deployments_file, out_dir, save=True):
    """Create a WebbPSF adjustable OTE object representing a perturbed OTE
    mirror state using mirror deployments defined in a YAML file.

    Parameters
    ----------
    deployments_file : str
        Path to YAML file containing deployments data. Expects format as
        produced by ``mirage.psf.deployments.generate_deployment_errors``
    out_dir : str
        Path to directory in which to save OPD FITS files
    save : bool, optional
        Denotes whether to save out the OPD (with and without tip/tilt)
        as FITs files

    Returns
    -------
    ote : webbpsf.opds.OTE_Linear_Model_WSS object
        WebbPSF OTE object describing perturbed OTE state with tip and tilt removed
    segment_tilts : numpy.ndarray
        List of X and Y tilts for each mirror segment, in microradians
    ote_opd_with_tilts : numpy.ndarray
        Array representing perturbed OTE OPD pupil before tip/tilt is removed
    """
    # Make an adjustable OTE object with WebbPSF
    nc = webbpsf.NIRCam()
    nc, ote = webbpsf.enable_adjustable_ote(nc)

    # Open existing file with previous deployments
    with open(deployments_file) as f:
        deployment_errors = yaml.unsafe_load(f)

    # Create OTE object and list of segment tilts
    ote, segment_tilts = apply_deployment_errors(
        ote, deployment_errors, out_dir=out_dir, save=save
    )
    ote_opd_with_tilts = ote.opd  # Save array to plot below

    # Remove tip and tilt
    ote = remove_piston_tip_tilt(ote, out_dir=out_dir, save=save)

    return ote, segment_tilts, ote_opd_with_tilts


def generate_random_ote_deployment(out_dir, reduction_factor=0.2, save=True):
    """Create a WebbPSF adjustable OTE object representing a perturbed OTE
    mirror state by randomly generating mirror deployment errors.

    Parameters
    ----------
    out_dir : str
        Path to directory in which to save OPD FITS files and the deployment
        error dictionary
    reduction_factor : optional
        Factor by which to reduce the input deployment errors. Default is 0.2.
    save : bool, optional
        Denotes whether to save out the OPD (with and without tip/tilt)
        as FITs files and the deployment error dictionary as a yaml

    Returns
    -------
    ote : webbpsf.opds.OTE_Linear_Model_WSS object
        WebbPSF OTE object describing perturbed OTE state with tip and tilt removed
    segment_tilts : numpy.ndarray
        List of X and Y tilts for each mirror segment, in microradians
    ote_opd_with_tilts : numpy.ndarray
        Array representing perturbed OTE OPD pupil before tip/tilt is removed
    """
    # Make an adjustable OTE object with WebbPSF
    nc = webbpsf.NIRCam()
    nc, ote = webbpsf.enable_adjustable_ote(nc)

    # Generate OPD and vector list with reduced deployment errors
    deployment_errors = generate_deployment_errors(out_dir=out_dir, save=save)
    deployment_errors = reduce_deployment_errors(
        deployment_errors, reduction_factor=reduction_factor, out_dir=out_dir, save=save
    )

    # Create an OTE object with those deployment errors
    ote, segment_tilts = apply_deployment_errors(
        ote, deployment_errors, out_dir=out_dir, save=save
    )
    ote_opd_with_tilts = ote.opd  # Save array to plot below

    # Remove tip and tilt
    ote = remove_piston_tip_tilt(ote, out_dir=out_dir, save=save)

    return ote, segment_tilts, ote_opd_with_tilts


def generate_deployment_errors(save=True, out_dir=None):
    """Randomly generate a expected deployment tolerances.

    Parameters:
    -----------
    save : bool, optional
        Denotes whether to save out the deployment error dictionary as a yaml
    out_dir : str, optional
        Directory in which to store the saved yaml file

    Returns:
    --------
    deployment_errors : dict
        Dictionary containing lists of generated deployment errors

    Notes:
    ------
    Deployment tolerances taken from JWST WFS&C Commissioning and Operations Plan (OTE-24):
    D36168 / 2299462 Rev C Page 10
    """

    deployment_errors = {
        'sm_piston': np.random.normal(loc=0, scale=2500/5),  # microns
        'sm_tilt': np.random.normal(loc=0, scale=1300/5, size=2),  # microradians
        'sm_decenter': np.random.normal(loc=0, scale=2500/5, size=2),  # microns
        'pm_piston': np.random.normal(loc=0, scale=1500/5, size=18),  # microns
        'pm_tilt': np.random.normal(loc=0, scale=1100/5, size=(18, 2)),  # microradians
        'pm_decenter': np.random.normal(loc=0, scale=1300/5, size=(18, 2)),  # microns
        'pm_roc': np.random.normal(loc=0, scale=151/5, size=(18)),  # microns
        'pm_clocking': np.random.normal(loc=0, scale=1200/5, size=(18)),  # microradians
        'global_pm_piston': np.random.normal(loc=0, scale=700/5),  # microns
        'global_pm_tilt': np.random.normal(loc=0, scale=190/5, size=2),  # microradians
        'global_pm_decenter': np.random.normal(loc=0, scale=200/5, size=2),  # microns
        'global_pm_clocking': np.random.normal(loc=0, scale=150/5),  # microradians
    }

    # Save the deployments dictionary to a yaml file that can be opened
    if out_dir is not None and save:
        save_file = os.path.join(out_dir, 'deployment_errors_{}.yaml'.format(time.strftime("%Y%m%d_%H%M%S")))
        with open(save_file, 'w') as f:
            yaml.dump(deployment_errors, f, default_flow_style=False)
        print('Saved deployment errors to {}'.format(save_file))
    elif save:
        raise IOError('Cannot save deployment errors to yaml; no out_dir provided')

    return deployment_errors

def reduce_deployment_errors(deployment_errors, reduction_factor=0.2, save=True, out_dir=None):
    """Reduce an existing dictionary of deployment errors by a given factor.

    Parameters:
    -----------
    deployment_errors : dict
        Dictionary containing lists of deployment errors to reduce
    reduction_factor : optional
        Factor by which to reduce the input deployment errors. Default is 0.2.
    save : bool, optional
        Denotes whether to save out the deployment error dictionary as a yaml
    out_dir : str, optional
        Directory in which to store the saved yaml file

    Returns:
    --------
    deployment_errors : dict
        Dictionary containing lists of reduced deployment errors
    """
    for key, value in deployment_errors.items():
        deployment_errors[key] = value * reduction_factor

    # Save the deployments dictionary to a yaml file that can be opened
    if out_dir is not None and save:
        save_file = os.path.join(out_dir, 'deployment_errors_reduced_{}.yaml'.format(time.strftime("%Y%m%d_%H%M%S")))
        with open(save_file, 'w') as f:
            yaml.dump(deployment_errors, f, default_flow_style=False)
        print('Saved reduced ({}%) deployment errors to {}'.format(reduction_factor * 100, save_file))
    elif save:
        raise IOError('Cannot save deployment errors to yaml; no out_dir provided')

    return deployment_errors


def apply_deployment_errors(ote, deployment_errors, save=True, out_dir=None):
    """Generate an OPD with the defined deployment errors.

    Parameters:
    -----------
    ote : webbpsf.opds.OTE_Linear_Model_WSS object
        Adjustable OTE object upon which to apply the deployment errors
    deployment_errors : dict
        Dictionary containing lists of deployment errors to reduce
    save : bool, optional
        Denotes whether to save out the deployment error dictionary as a yaml
    out_dir : str, optional
        Directory in which to store the saved yaml file

    Returns:
    --------
    ote : webbpsf.opds.OTE_Linear_Model_WSS object
        Adjustable OTE object upon which deployment errors were applied
    segment_tilts : np.ndarray
        List of X & Y tilts for each segment, in microns, to be used later to
        calculate the segment PSF displacement in pixels
    """
    ote.reset()
    ote.remove_piston_tip_tilt = False  # Reset the piston/tip/tilt

    # Add SM moves
    ote.move_sm_local(
        piston=deployment_errors['sm_piston'], xtilt=deployment_errors['sm_tilt'][0],
        ytilt=deployment_errors['sm_tilt'][1], xtrans=deployment_errors['sm_decenter'][0],
        ytrans=deployment_errors['sm_decenter'][1])

    # Add PMSA (segment) moves
    for i, seg in enumerate(ote.segnames[0:18]):
        ote.move_seg_local(
            seg,
            piston=deployment_errors['pm_piston'][i] + deployment_errors['global_pm_piston'],
            xtilt=deployment_errors['pm_tilt'][i][0] + deployment_errors['global_pm_tilt'][0],
            ytilt=deployment_errors['pm_tilt'][i][1] + deployment_errors['global_pm_tilt'][1],
            xtrans=deployment_errors['pm_decenter'][i][0] + deployment_errors['global_pm_decenter'][0],
            ytrans=deployment_errors['pm_decenter'][i][1] + deployment_errors['global_pm_decenter'][1],
            roc=deployment_errors['pm_roc'][i],
            clocking=deployment_errors['pm_clocking'][i] + deployment_errors['global_pm_clocking']
        )

    # Save segment tilt to a list
    segment_tilts = np.copy(ote.segment_state[:, :2])

    if out_dir is not None and save:
        save_file = os.path.join(out_dir, 'OPD_withtilt_{}.fits'.format(time.strftime("%Y%m%d_%H%M%S")))
        hdu = fits.PrimaryHDU(ote.opd, header=ote.opd_header)
        hdu.writeto(save_file)
        print('Saved OPD to {}'.format(save_file))
    elif save:
        raise IOError('Cannot save deployment errors to yaml; no out_dir provided')

    return ote, segment_tilts


def remove_piston_tip_tilt(ote, save=True, out_dir=None):
    """Remove the piston/tip/tilt from the OPD.

    Parameters:
    -----------
    ote : webbpsf.opds.OTE_Linear_Model_WSS object
        Adjustable OTE object from which to remove piston/tip/tilt\
    save : bool, optional
        Denotes whether to save out the deployment error dictionary as a yaml
    out_dir : str, optional
        Directory in which to store the saved yaml file

    Returns:
    --------
    ote : webbpsf.opds.OTE_Linear_Model_WSS object
        Adjustable OTE object without piston/tip/tilt
    """
    ote.remove_piston_tip_tilt = True

    # Also manually zero out the piston/tip/tilt
    for segment in range(18):
        ote.segment_state[segment, :3] = 0
    ote.update_opd()

    if out_dir is not None and save:
        save_file = os.path.join(out_dir, 'OPD_notilt_{}.fits'.format(time.strftime("%Y%m%d_%H%M%S")))
        hdu = fits.PrimaryHDU(ote.opd, header=ote.opd_header)
        hdu.writeto(save_file)
        print('Saved OPD with piston/tip/tilt removed to {}'.format(save_file))
    elif save:
        raise IOError('Cannot save deployment errors to yaml; no out_dir provided')

    return ote



#! /usr/bin/env python

"""
Tools for working with PSFs
"""
import os

from astropy.io import fits
from astropy.modeling import models, fitting
from astropy.modeling.models import Gaussian2D
import numpy as np
from photutils.datasets import load_irac_psf
import webbpsf

from mirage.psf.psf_selection import get_psf_wings
from mirage.utils.siaf_interface import get_instance


def gaussian_psf(fwhm, xdim, ydim):
    """Return a 2D Gaussian kernel

    Parameters
    ----------
    fwhm : float
        FWHM of the Gaussian kernel to return

    xdim : int
        Length of the created PSF in the x-dimension

    ydim : int
        Length of the created PSF in the y-dimension

    Returns
    -------
    kernel : numpy.ndarray
        2D array containing 2D Gaussian kernel
    """
    # Center the PSF in the array
    xmean = xdim // 2
    ymean = ydim // 2

    # Translate FWHM to standard deviation
    dev = fwhm / 2.35

    # Create model
    model = Gaussian2D(amplitude=1., x_mean=xmean, y_mean=ymean, x_stddev=dev, y_stddev=dev)

    # Evalulate model
    x_grid, y_grid = np.mgrid[0: xdim, 0: ydim]
    kernel = model.evaluate(x=x_grid, y=y_grid, amplitude=1., x_mean=xmean,
                            y_mean=ymean, x_stddev=dev, y_stddev=dev, theta=0.)
    return kernel


def get_HST_PSF(fwhm_pixels):
    """Get a representative PSF for the HST instrument/detector in
    question. Currently this is just a 2D Gaussian of the appropriate
    FWHM.

    Parameters
    ----------
    fwhm_pixels : float
        FWHM of the Gaussian in units of pixels

    Returns
    -------
    psf : numpy.ndarray
        2D array containing PSF

    """
    # Set the dimensions of the PSF array to be +/- 50 * FWHM
    dim = np.int(np.round(100 * fwhm_pixels))

    # Make sure the array has an odd number of rows and columns
    if dim % 2 == 0:
        dim += 1

    # Create 2D Gaussian
    return gaussian_psf(fwhm_pixels, dim, dim)


def get_IRAC_PSF(meta):
    """Use photutils to get a representative PSF for the IRAC data
    specified in ``meta``.

    Parameters
    ----------
    meta : dict
        Dictionary of instrument information. Most important is
        meta['channel'] which contains the integer IRAC channel
        number. PSFs are loaded from photutils

    Returns
    -------
    hdu.data : numpy.ndarray
        2D array containing the IRAC PSF
    """
    if isinstance(meta['channel'], int):
        if meta['channel'] >= 1 and meta['channel'] <= 4:
            hdu = load_irac_psf(channel=meta['channel'])
        else:
            raise ValueError("ERROR: IRAC Channel number must be an integer from 1 to 4.")
    else:
        raise ValueError("ERROR: IRAC Channel number must be an integer from 1 to 4.")

    pixscale = hdu.header['SECPIX']
    mosaic_pix_scale1 = meta['pix_scale1']
    if not np.isclose(mosaic_pix_scale1, pixscale, rtol=0.05):
        raise ValueError(("Reported pixel scale in the mosaic does not match that in "
                          "the IRAC PSF from photutils. If the pixel scale of the mosaic "
                          "is really not {}, please provide your own fits file containing "
                          "a representative PSF at the desired pixel scale.".format(pixscale)))

    return hdu.data


def get_JWST_pixel_scale(meta, aperture=None):
    """Use pysiaf to find the nominal pixel scale for a given JWST
    instrument/detector

    Parameters
    ----------
    meta : dict
        Dictionary of instrument information. Most important is
        meta['channel'] which contains the integer IRAC channel
        number. PSFs are loaded from photutils

    aperture : str
        Name of aperture to look up pixel scale for. If None,
        then fall back to using FULL or CEN as appropriate.

    Returns
    -------
    siaf_aper.XSciScale : float
        Pixel scale (arcsec/pixel) in x-direction on detector

    siaf_aper.YSciScale : float
        Pixel scale (arcsec/pixel) in y-direction on detector
    """
    # Create SIAF instance
    siaf = get_instance(meta['instrument'])

    # If needed, get the name to append to the detector to create
    # the full aperture
    # name
    if aperture is None:
        if meta['instrument'].lower() == 'nircam':
            aperture = 'FULL'
        elif meta['instrument'].lower() == 'niriss':
            aperture = 'CEN'
        elif meta['instrument'].lower() == 'fgs':
            aperture = 'FULL'

        # In SIAF, the FGS detectors are FGS1,2 rather than GUIDER1,2
        detector = meta['detector'].upper()
        if 'GUIDER' in detector:
            detector = detector.replace('GUIDER', 'FGS')

        aperture = '{}_{}'.format(detector, aperture)

    # Get the aperture-specific SIAF info
    siaf_aper = siaf[aperture]

    # Return the x and y pixel scale
    return siaf_aper.XSciScale, siaf_aper.YSciScale


def get_JWST_PSF(meta):
    """Use webbpsf to get a representative PSF for the JWST data specified
    in ``meta``.

    Parameters
    ----------
    meta : dict
        Dictionary of instrument information.

    Returns
    -------
    psf_model : numpy.ndarray
        2D array containing the PSF
    """
    # Find the nominal pixel scale, which is what the psf_wings
    # library is saved at
    nominal_pix_scale_x, nominal_pix_scale_y = get_JWST_pixel_scale(meta)

    mosaic_pix_scale1 = meta['pix_scale1']
    if not np.isclose(mosaic_pix_scale1, nominal_pix_scale_x, rtol=0.05):
        raise ValueError(("Reported pixel scale in the mosaic does not match the nominal "
                          "pixel scale for {} {}/{}. If the pixel scale of the mosaic "
                          "is really not {}, please provide your own fits file containing "
                          "a representative PSF at the desired pixel scale.".format(meta['instrument'],
                                                                                    meta['filter'],
                                                                                    meta['pupil'],
                                                                                    nominal_pix_scale_x)))
    # If the pixel scale reported in the mosaic file is the same as the
    # nominal pixel scale for that instrument/detector, then just read in
    # the PSF from the psf_wings file
    instrument = meta['instrument'].lower()
    psf_model = get_psf_wings(meta['instrument'], meta['detector'], meta['filter'], meta['pupil'],
                              'predicted', 0, os.path.join(os.path.expandvars('$MIRAGE_DATA'),
                                                           instrument, 'gridded_psf_library/psf_wings'))

    return psf_model


def get_psf_metadata(filename):
    """Retrieve basic PSF-related metadata

    Parameters
    ----------
    filename : str
        Name of fits file containing mosaic

    Returns
    -------
    metadata : dict
        Dictionary of instrument-specific data
    """
    metadata = {}
    with fits.open(filename) as hdulist:

        # We want basic metadata on the instrument. Assume it's in the
        # primary header
        header = hdulist[0].header

        metadata['telescope'] = metadata_entry('TELESCOP', header).upper()
        metadata['instrument'] = metadata_entry('INSTRUME', header).upper()

        if metadata['telescope'] == 'JWST':
            metadata['detector'] = metadata_entry('DETECTOR', header)
            metadata['filter'] = metadata_entry('FILTER', header)
            metadata['pupil'] = metadata_entry('PUPIL', header)
            metadata['pix_scale1'] = np.abs(metadata_entry('CD1_1', header)) * 3600.
            metadata['pix_scale2'] = np.abs(metadata_entry('CD2_2', header)) * 3600.

        if metadata['telescope'] == 'HST':
            metadata['detector'] = metadata_entry('DETECTOR', header)
            metadata['filter'] = metadata_entry('FILTER', header)
            metadata['pa'] = metadata_entry('PA_APER', header)  # PA of reference aperture center
            metadata['pix_scale1'] = np.abs(metadata_entry('CD1_1', header)) * 3600.
            metadata['pix_scale2'] = np.abs(metadata_entry('CD2_2', header)) * 3600.

        if metadata['instrument'] == 'IRAC':
            metadata['channel'] = int(metadata_entry('CHNLNUM', header))
            metadata['pix_scale1'] = np.abs(metadata_entry('PXSCAL1', header))
            metadata['pix_scale2'] = np.abs(metadata_entry('PXSCAL2', header))
            metadata['pa'] = metadata_entry('PA', header)  # deg] Position angle of axis 2 (E of N)

        if metadata['telescope'] not in 'JWST HST SPITZER'.split():
            metadata['pix_scale1'] = np.abs(metadata_entry('CD1_1', header)) * 3600.
            metadata['pix_scale2'] = np.abs(metadata_entry('CD2_2', header)) * 3600.

    return metadata


def measure_fwhm(array):
    """Fit a Gaussian2D model to a PSF and return the FWHM

    Parameters
    ----------
    array : numpy.ndarray
        Array containing PSF

    Returns
    -------
    x_fwhm : float
        FWHM in x direction in units of pixels

    y_fwhm : float
        FWHM in y direction in units of pixels
    """
    yp, xp = array.shape
    y, x, = np.mgrid[:yp, :xp]
    p_init = models.Gaussian2D()
    fit_p = fitting.LevMarLSQFitter()
    fitted_psf = fit_p(p_init, x, y, array)
    return fitted_psf.x_fwhm, fitted_psf.y_fwhm


def metadata_entry(keyword, header):
    """Get the ``header`` keyword value from the ``header`` object. If
    it is not present, return None

    Paramters
    ---------
    keyword : str
        Name of header keyword to examine

    header : astropy.io.fits.header
        Header object from fits file

    Returns
    -------
    value : str, int, float
        Value of the header keyword
    """
    try:
        value = header[keyword]
    except KeyError:
        value = None
    return value


def same_array_size(array1, array2):
    """Crop the input arrays such that they are the same size in both
    dimensions

    Parameters
    ----------
    array1 : numpy.ndarray
        2D array

    array2 : numpy.ndarray
        2D array

    Returns
    -------
    array1 : numpy.ndarray
        Potentially cropped 2D array

    array2 : numpy.ndarray
        Potentially cropped 2D array

    """
    shape1 = array1.shape
    shape2 = array2.shape
    miny = min([shape1[0], shape2[0]])
    minx = min([shape1[1], shape2[1]])

    # Crop y dimension
    if shape1[0] == miny:
        smaller = array1
        larger = array2
        smaller_shape = shape1
        larger_shape = shape2
    else:
        smaller = array2
        larger = array1
        smaller_shape = shape2
        larger_shape = shape1

    diff = abs(larger_shape[0] - miny)
    low_delta = diff // 2
    if diff % 2 == 1:
        high_delta = low_delta + 1
    else:
        high_delta = low_delta
    larger = larger[low_delta: larger_shape[0]-high_delta, :]

    if shape1[0] == miny:
        array1 = smaller
        array2 = larger
    else:
        array1 = larger
        array2 = smaller
    shape1 = array1.shape
    shape2 = array2.shape

    # Crop x dimension
    if shape1[1] == minx:
        smaller = array1
        larger = array2
        smaller_shape = shape1
        larger_shape = shape2
    else:
        smaller = array2
        larger = array1
        smaller_shape = shape2
        larger_shape = shape1

    diff = abs(larger_shape[1] - minx)
    low_delta = diff // 2
    if diff % 2 == 1:
        high_delta = low_delta + 1
    else:
        high_delta = low_delta
    larger = larger[:, low_delta: larger_shape[1]-high_delta]

    if shape1[1] == minx:
        array1 = smaller
        array2 = larger
    else:
        array1 = larger
        array2 = smaller
    shape1 = array1.shape
    shape2 = array2.shape

    return array1, array2



#! /usr/bin/env python

'''
Class that creates an integration containing multiple frames and
shows a source that is moving relative to the detector.

Arguments:
----------
stamp -- 2D stamp image containing target
xframes -- list of x-coordinate pixel position of target
           in each frame
yframes -- list of y-coordinate pixel position of target
           in each frame
frametime -- exposure time in seconds corresponding to one
             detector readout (varies with subarray size)
outx -- x-dimension size of the output aperture (2048 for
        full-frame)
outy -- y-dimension size of the output aperture (2048 for
        full-frame)

Returns:
--------
3D array containing the signal of the source in each frame
of the integration

Author:
-------
Bryan Hilbert
'''

import sys

import numpy as np
from astropy.io import fits

class MovingTarget():

    def __init__(self):
        self.verbose = False
        self.subsampx = 3
        self.subsampy = 3

    def create(self, stamp, xframes, yframes, frametime, outx, outy):
        """
        MAIN FUNCTION

        Arguments:
        ----------
        stamp -- 2D stamp image containing target
        xframes -- list of x-coordinate pixel position of target
                   in each frame
        yframes -- list of y-coordinate pixel position of target
                   in each frame
        frametime -- exposure time in seconds corresponding to one
                     detector readout (varies with subarray size)
        outx -- x-dimension size of the output aperture (2048 for
                full-frame)
        outy -- y-dimension size of the output aperture (2048 for
                full-frame)

        Returns:
        --------
        3D array containing the signal of the source in each frame
        of the integration
        """

        # Make sure subsampling factor is an integer
        self.subsampx = np.int(self.subsampx)
        self.subsampy = np.int(self.subsampy)

        # Quick fix for the case where xinit,yinit are integers
        xinit = np.float(xframes[1])
        yinit = np.float(yframes[1])

        # Change position angle to radians
        #posang = posang * np.pi / 180.

        # List of times for all frames
        numframes = len(xframes)-1
        times = frametime * np.arange(-1, numframes)

        # Generate a list of locations at dist-pixel increments
        # between the beginning and ending locations
        xs, ys = self.equidistantXY(xframes[0], yframes[0], xframes[-1],
                                    yframes[-1], 1./self.subsampx)

        # Subsample the PSF image
        substamp = self.subsample(stamp, self.subsampx, self.subsampy)
        substamplen = substamp.shape

        # Create the initial output frame
        ystamplen, xstamplen = stamp.shape
        #minx = np.min([np.floor(xinit) - np.ceil(xstamplen/2.),np.floor(xframes[-1])-np.ceil(xstamplen/2.)])
        minx = np.int(np.min([np.floor(xframes[0]) - np.ceil(xstamplen/2.),\
                              np.floor(xframes[-1])-np.ceil(xstamplen/2.)]))
        maxx = np.int(np.max([np.floor(xframes[-1] + xstamplen/2.),\
                              np.floor(xframes[0]+xstamplen/2.)]))
        #miny = np.min([np.floor(yinit) - np.ceil(ystamplen/2.),np.floor(yframes[-1])-np.ceil(ystamplen/2.)])
        miny = np.int(np.min([np.floor(yframes[0]) - np.ceil(ystamplen/2.),\
                              np.floor(yframes[-1])-np.ceil(ystamplen/2.)]))
        maxy = np.int(np.max([np.floor(yframes[-1] + ystamplen/2.),\
                              np.floor(yframes[0]+ystamplen/2.)]))

        # Don't let stamps fall off the edges of the output array
        mnx = minx
        mxx = maxx
        mny = miny
        mxy = maxy
        if minx < 0:
            mnx = 0
        if maxx > outx:
            mxx = np.int(outx-1)
        if miny < 0:
            mny = 0
        if maxy > outy:
            mxy = np.int(outy-1)

        # Subsample the output frame
        totxpoints = np.min([outx,mxx-mnx+1])
        totypoints = np.min([outy,mxy-mny+1])
        outputframe0 = np.zeros((np.int(totypoints*self.subsampy),\
                                 np.int(totxpoints*self.subsampx)))
        outputframe1 = np.zeros((np.int(totypoints*self.subsampy),\
                                 np.int(totxpoints*self.subsampx)))
        outfull = np.zeros((numframes, outy, outx))
        outsubshape = outputframe0.shape

        # Translate the source location x and y values to the coordinates
        # of the output frame
        deltacenterx = np.round(self.subsampx / 2. - 1 + 0.000001)
        deltacentery = np.round(self.subsampy / 2. - 1 + 0.000001)
        xframessub = np.round((xframes-mnx) * self.subsampx) + deltacenterx
        yframessub = np.round((yframes-mny) * self.subsampy) + deltacentery
        xssub = np.round((xs-mnx) * self.subsampx) + deltacenterx
        yssub = np.round((ys-mny) * self.subsampy) + deltacentery

        for i in range(1,numframes+1):
            # Find the velocity of the source during this frame
            xvelocity = (xframes[i] - xframes[i-1]) / frametime
            yvelocity = (yframes[i] - yframes[i-1]) / frametime
            secPerPix = 1. / np.sqrt(xvelocity*xvelocity + yvelocity*yvelocity)
            outputframe1 = np.copy(outputframe0)

            if xframessub[i-1] < xframessub[i]:
                goodxs = ((xssub > (xframessub[i-1]+1e-7)) & (xssub < (xframessub[i]-1e-7)))
            else:
                goodxs = ((xssub > (xframessub[i]+1e-7)) & (xssub < (xframessub[i-1]-1e-7)))
            if yframessub[i-1] < yframessub[i]:
                goodys = ((yssub > (yframessub[i-1]+1e-7)) & (yssub < (yframessub[i]-1e-7)))
            else:
                goodys = ((yssub > (yframessub[i]+1e-7)) & (yssub < (yframessub[i-1]-1e-7)))

            xsum = np.sum(goodxs)
            ysum = np.sum(goodys)
            if xsum >= ysum:
                good = goodxs
            else:
                good = goodys

            if (np.all((xframes[i-1:i+1]-xstamplen) > outx) or \
                (np.all((yframes[i-1:i+1]-ystamplen) > outy))):
                outputframe1 = np.copy(outputframe0)
            elif (np.all((xframes[i-1:i+1]+xstamplen) < 0) or \
                  (np.all((yframes[i-1:i+1]+ystamplen) < 0))):
                outputframe1 = np.copy(outputframe0)
            else:
                outputframe1 = self.inputMotion(outputframe1, substamp, xframessub[i-1:i+1],
                                                yframessub[i-1:i+1],xssub[good],yssub[good],
                                                secPerPix)

            outputframe0 = np.copy(outputframe1)

            # Put the output frames back to the original resolution
            resampled = self.resample(outputframe1, self.subsampx, self.subsampy)
            resampylen, resampxlen = resampled.shape
            #outfull[i-1,mny:mxy+1,mnx:mxx+1] = resampled
            maxfully = mny + resampylen
            maxfullx = mnx + resampxlen
            maxrey = resampylen
            maxrex = resampxlen
            if (mny + resampylen) > outy:
                diffind = (mny + resampylen) - outy
                maxfully = outy
                maxrey -= diffind
            if (mnx + resampxlen) > outx:
                diffind = (mnx + resampxlen) - outx
                maxfullx = outx
                maxrex -= diffind
            outfull[i-1, mny:maxfully, mnx:maxfullx] = resampled[0:maxrey, 0:maxrex]
        return outfull

    def resample(self, frame, sampx, sampy):
        """
        Return subsampled image back to original resolution

        Arguments:
        ----------
        frame -- subsampled image
        sampx -- x-dimension subsampling factor (e.g. 3 means 3x oversampled
                 compared to original image
        sampy -- y-dimension subsampling factor

        Returns:
        --------
        resampled image
        """
        framey, framex = frame.shape
        newframe = np.zeros((np.int(framey/sampy), np.int(framex/sampx)))
        newframey, newframex = newframe.shape

        for j in range(newframey):
            for i in range(newframex):
                newframe[j, i] = np.sum(frame[sampy*j:sampy*(j+1), sampx*i:sampx*(i+1)])
        return newframe

    def coordCheck(self, center, len_stamp, len_out):
        """
        Find indexes of stamp and frame to use
        given that the stamp may fall off the edge
        of the frame. Works on only one coordinate dimension

        Arguments:
        ----------
        center -- coordinate (in full aperture coords of the center of the
                  stamp image.
        len_stamp -- Size of the stamp image
        len_out -- Size of the full aperture image

        Returns:
        --------
        x and y coordinates corresponding to the beginning and ending
        (i.e. top and bottom for y-dimension, or left and right for x-
        dimension) of the stamp image on the full frame aperture, as
        well as the beginning and ending coordinates within the stamp
        image that fall onto the full frame aperture.
        """
        outxmin = center - len_stamp/2
        outxmax = outxmin + len_stamp
        stampxmin = 0
        stampxmax = len_stamp

        # Left edge of stamp is off the edge of output
        if outxmin < 0:
            if outxmin >= (0.-len_stamp):
                stampxmin = 0 - outxmin
                outxmin = 0
            else:
                # Here the image is completely off the output frame
                stampxmin = np.nan
                outxmin = np.nan
                stampxmax = np.nan
                outxmax = np.nan

        # Right edge of stamp is off the edge of the output
        if outxmax > len_out:
            if outxmax <= (len_out + len_stamp):
                delta = outxmax - len_out
                stampxmax = len_stamp - delta
                outxmax = len_out
            else:
                # Here the image is completely off the left side of output
                outxmax = np.nan
                stampxmax = np.nan
                outxmin = np.nan
                stampxmin = np.nan

        indexes = [outxmin, outxmax, stampxmin, stampxmax]
        if np.all(np.isfinite(indexes)):
            ioutxmin = np.int(outxmin)
            ioutxmax = np.int(outxmax)
            istampxmin = np.int(stampxmin)
            istampxmax = np.int(stampxmax)
            dout = ioutxmax - ioutxmin
            dstamp = istampxmax - istampxmin

            if dout == dstamp:
                pass
            elif dout == (dstamp+1):
                if istampxmin > 0:
                    istampxmin -= 1
                else:
                    istampxmax += 1
            elif dstamp == (dout+1):
                if ioutxmin > 0:
                    ioutxmin -= 1
                else:
                    ioutxmax += 1
            else:
                print("WARNING: bad stamp/output match. Quitting.")
                sys.exit()
            return ioutxmin, ioutxmax, istampxmin, istampxmax
        else:
            # If values are NaN then we can't change them to integers
            return outxmin, outxmax, stampxmin, stampxmax

    def inputMotion(self, inframe, source, xbounds, ybounds, xs, ys, secperpix):
        """
        Smear out the source to create an output frame image
        given the necessary info about the source location and velocity

        Arguments:
        ----------
        inframe -- 2D array representing the image
        source -- 2D stamp image containing the source
        xbounds -- 2-element list containing the starting and ending x-dimension
                   coordinates of the source (i.e. location corresponding to the
                   beginning and ending of the frame)
        ybounds -- 2-element list containing the starting and ending y-dimension
                   coordinates of the source
        xs -- list of x-coordinate positions of the source
        ys -- list of y-coordinate positions of the source
        secperpix -- Inverse velocity of the source, in seconds per pixel
        """
        frameylen,framexlen = inframe.shape
        srcylen,srcxlen = source.shape
        xlist = np.append(xs, xbounds[1])
        ylist = np.append(ys, ybounds[1])
        xlist = np.insert(xlist, 0, xbounds[0])
        ylist = np.insert(ylist, 0, ybounds[0])
        xlist = np.round(xlist)
        ylist = np.round(ylist)

        for i in range(1,len(xlist)):
            outxmin, outxmax, stampxmin, stampxmax = self.coordCheck(xlist[i], srcxlen, framexlen)
            outymin, outymax, stampymin, stampymax = self.coordCheck(ylist[i], srcylen, frameylen)
            outcoords = np.array([outxmin,outxmax,outymin,outymax])

            # If any of the coordinates are set to NaN, then the stamp image is completely off
            # the output frame and it shouldn't be added
            if np.all(np.isfinite(outcoords)):
                dist = np.sqrt((xlist[i]-xlist[i-1])**2 + (ylist[i]-ylist[i-1])**2)
                inframe[outymin:outymax, outxmin:outxmax] += (source[stampymin:stampymax, stampxmin:stampxmax]*secperpix*dist)
        return inframe

    def subsample(self, image, factorx, factory):
        """
        Subsample the input image

        Arguments:
        ----------
        image -- 2D image
        factorx -- factor in the x-dimension to subsample the image
                 (e.g. factorx=2 will break each pixel into 2 pixels
                  in the x dimension)
        factory -- factor in the y-dimension to subsample the image

        Setting factorx = 2, factory = 2 will break each pixel in the
        original image into a 2x2 grid of pixels

        Returns:
        --------
        Subsampled image
        """
        ydim, xdim = image.shape
        substamp = np.zeros((ydim*factory, xdim*factorx))

        for i in range(xdim):
            for j in range(ydim):
                substamp[factory*j:factory*(j+1), factorx*i:factorx*(i+1)] = image[j, i]
        return substamp

    def equidistantXY(self,xstart, ystart, xend, yend, dist):
        """
        Return a list of x,y positions that are equidistant
        between the beginning and ending positions, with
        a distance of dist pixels between them

        Arguments:
        ----------
        xstart -- beginning x coordinate
        ystart -- beginning y coordinate
        xend -- ending x coordinate
        yend -- ending y coordinate
        dist -- distance in pixels between adjacent positions
        """
        xlen = 0
        ylen = 0

        deltax = xend - xstart
        deltay = yend - ystart
        ang = np.arctan2(deltay, deltax)

        dx = np.cos(ang) * dist
        dy = np.sin(ang) * dist

        if dx != 0.:
            xs = np.arange(xstart, xend+dx/2, dx)
            xlen = len(xs)
        else:
            # Motion parallel to y axis
            xs = np.array([xstart])

        if dy != 0:
            ys = np.arange(ystart, yend+dy/2, dy)
            ylen = len(ys)
        else:
            # Motion parallel to x asis
            ys = np.array([ystart])

        # Make sure lengths agree
        if xlen == 0:
            xs = np.zeros(ylen) + xstart
        if ylen == 0:
            ys = np.zeros(xlen) + ystart

        return xs, ys

    def radecPerFrame(self, ra0, dec0, ravel, decvel, time):
        """
        Generate a list of RA,Dec locations for a source in
        a series of frames

        Arguments:
        ----------
        ra0 -- Initial RA value of source
        dec0 -- Initial Dec value of source
        ravel -- Velocity of source in the RA direction
        decvel -- Velocity of source in the Dec direction
        time -- List of times corresponding to all frames

        Returns:
        --------
        List of RA, Dec positions corresponding to all input times
        """
        ra = ra0 + (ravel * time)
        dec = dec0 + (decvel * time)
        return ra, dec

    def xyPerFrame(self, velocity, time, ang, x0, y0):
        """
        Generate list of x,y positions for a source given an
        initial position, velocity, and velocity angle

        Arguments:
        ----------
        velocity -- Velocity of source
        time -- List of times at which we want to find positions
        ang -- Angle (in radians) at which source is traveling.
               An ang of zero corresponds to moving along the +x axis.
               An ang of np.pi/2 corresponds to moveing along the +y axis.
        x0 -- Initial x coordinate of source
        y0 -- Initial y coordinate of source

        Returns:
        --------
        Tuple of x-list, y-list souce locations
        """
        ratex = velocity * np.cos(ang)
        ratey = velocity * np.sin(ang)

        # x,y in each frame
        xs = x0 + ratex*time
        ys = y0 + ratey*time

        return xs,ys



#! /usr/bin/env python
"""Test the file name generation in terms of formatting.

Authors
-------
    - Johannes Sahlmann

Use
---
    >>> pytest -s test_filename_generation.py


"""

import glob
import os

from astropy.table import Table
import numpy as np
import pytest

from mirage.yaml import generate_observationlist, yaml_generator
from .test_apt_interface import ON_TRAVIS, temporary_directory, TEST_DATA_DIR

@pytest.mark.skipif(ON_TRAVIS,
                   reason="Cannot access mirage data in the central storage directory from Travis CI.")
def test_filenames(temporary_directory):
    """Exercise mirage input generation from APT files (.xml and .pointing)."""

    for instrument in ['misc']:
        apt_dir = os.path.join(TEST_DATA_DIR, instrument)

        if instrument == 'misc':
            apt_file_seeds = ['12345/12345_nircam_imaging_prime_niriss_wfss_parallel.xml']
            source_list_file_name = os.path.join(apt_dir, 'niriss_point_sources.list')

        catalogs = {}
        for instrument_name in 'fgs nircam niriss miri nirspec'.split():
            if instrument_name.lower() == 'nircam':
                catalogs[instrument_name.lower()] = {}
                catalogs[instrument_name.lower()]['sw'] = source_list_file_name
                catalogs[instrument_name.lower()]['lw'] = source_list_file_name
            else:
                catalogs[instrument_name.lower()] = source_list_file_name

        for i, apt_file_seed in enumerate(apt_file_seeds):
            print('\n\n' + '=' * 100 + '\n')

            obs_yaml_files = glob.glob(os.path.join(temporary_directory, 'jw*.yaml'))
            for file in obs_yaml_files:
                os.remove(file)

            if '.xml' in apt_file_seed:
                apt_file_xml = os.path.join(apt_dir, apt_file_seed)
                apt_file_pointing = os.path.join(apt_dir, apt_file_seed.replace('.xml', '.pointing'))
                observation_list_file = os.path.join(temporary_directory,
                                                     '{}_observation_list.yaml'.format(apt_file_seed.replace('/', '_').split('.')[0]))


            print('Processing program {}'.format(apt_file_xml))

            yam = yaml_generator.SimInput(input_xml=apt_file_xml, pointing_file=apt_file_pointing,
                                          observation_list_file=observation_list_file,
                                          verbose=True, output_dir=temporary_directory, simdata_output_dir=temporary_directory,
                                          offline=True)
            try:
                yam.create_inputs()
            except RuntimeError as e:
                print('\nERROR detected. Skipping {} because of error\n{}\n\n'.format(apt_file_xml, e))
                continue


            Table([yam.info['yamlfile']]).pprint(max_lines=-1)

            exposure_numbers = np.array([np.int(s.split('_')[2]) for s in yam.info['yamlfile']])
            assert np.max(exposure_numbers) <= 18



#! /usr/bin/env python

"""
This module contains utilities used for reading and writing various files
used or created by Mirage.

Author
------

    - Bryan Hilbert

Use
---

    This module can be imported and used as such:

    ::

        from mirage.utils import file_io
        gain, gain_head = file_io.read_gain_file(filename)
"""
import os

from astropy.io import ascii, fits
import numpy as np


def read_filter_throughput(filename):
    """Read in the ascii file containing a filter throughput curve

    Parameters
    ----------
    filename : str
        Name of ascii file containing throughput info

    Returns
    -------
    (wavelengths, throughput) : tup
        Tuple of 1D numpy arrays containing the wavelength and throughput
        values from the file
    """
    tab = ascii.read(filename)
    return tab['Wavelength_microns'].data, tab['Throughput'].data


def read_gain_file(filename):
    """
    Read in CRDS-formatted gain reference file

    Paramters
    ---------
    filename : str
        Name of (fits) gain reference file

    Returns
    -------
    image : numpy.ndarray
        2D array gain map

    header : dict
        Information contained in the header of the file
    """
    try:
        with fits.open(filename) as h:
            image = h[1].data
            header = h[0].header
    except (FileNotFoundError, IOError) as e:
        print(e)

    mngain = np.nanmedian(image)

    # Set pixels with a gain value of 0 equal to mean
    image[image == 0] = mngain

    # Set any pixels with non-finite values equal to mean
    image[~np.isfinite(image)] = mngain
    return image, header



#! /usr/bin/env python

'''
Run the linearity correction step backwards on an input
image or ramp. Introduce non-linearity into a linear
input ramp.
'''

import sys
import numpy as np


def unlinearize(image, coeffs, sat, lin_satmap, maxiter=10, accuracy=0.000001, robberto=False,
                save_accuracy_map=False, accuracy_file='unlinearize_no_convergence.fits'):
    # Insert non-linearity into the linear synthetic sources

    # If the sizes of the satmap or coeffs are different than
    # the data, return an error
    if sat.shape != image.shape[-2:]:
        raise ValueError(("WARNING: image y,x shape is {}, but input saturation map shape is {}."
                         .format(image.shape[-2:], sat.shape)))

    # REWORK SO THAT THE LINARIZED SATURATION MAP IS AN INPUT
    # RATHER THAN BEING CREATED HERE. THIS IS BECAUSE THE SUPERBIAS
    # AND REFPIX SIGNALS MUST BE SUBTRACTED BEFORE LINEARIZING
    # THE ORIGINAL RAW SATURATION MAP
    # Sat is the original saturation map data for non-linear ramps.
    # Translate to saturation maps for the linear ramps here so
    # that we can pay attention only to non-saturated pixels in
    # the input linear image. Do this by applying the standard
    # linearity correction to the saturation map
    # lin_satmap = nonLinFunc(sat,coeffs,sat)

    # Find pixels with "good" signals, to have the nonlin applied.
    # Negative pix or pix with signals above the requested max
    # value will not be changed.
    x = np.copy(image)
    i1 = np.where((image > 0.) & (image < lin_satmap))
    dev = np.zeros_like(image, dtype=float)
    dev[i1] = 1.
    i2 = np.where((image <= 0.) | (image >= lin_satmap))
    numhigh = np.where(image >= lin_satmap)
    i = 0

    # Initial run of the nonlin function - when calling the
    # non-lin function, give the original satmap for the
    # non-linear signal values
    val = nonLinFunc(image, coeffs, sat)
    val[i2] = 1.

    if robberto:
        x = image * val
    else:
        x[i1] = (image[i1]+image[i1]/val[i1]) / 2.
        while i < maxiter:
            i = i + 1
            val = nonLinFunc(x, coeffs, sat)
            val[i2] = 1.
            dev[i1] = np.abs(image[i1] / val[i1]-1.)
            inds = np.where(dev[i1] > accuracy)
            if inds[0].size < 1:
                break
            val1 = nonLinDeriv(x, coeffs, sat)
            val1[i2] = 1.
            x[i1] = x[i1] + (image[i1] - val[i1]) / val1[i1]

    # If we max out the number of iterations,
    # save the array of accuracy values. Spot
    # checks reveal the pix that don't meet the
    # accuracy reqs are randomly located on the detector,
    # and don't seem to be correlated with point source
    # locations.
    if i == maxiter and save_accuracy_map:
        from astropy.io import fits
        print(("WARNING: some pixels failed to unlinearize correctly within "
               "the maximum number of iterations. Map of accuracy of the "
               "unlinearized values saved to {}.".format(accuracy_file)))
        devcheck = np.copy(dev)
        devcheck[i2] = -1.
        h0 = fits.PrimaryHDU()
        h1 = fits.ImageHDU(devcheck)
        hl = fits.HDUList([h0, h1])
        hl.writeto(accuracy_file, overwrite=True)
    return x


def nonLinFunc(image, coeffs, limits):
    # Apply linearity correction coefficients
    # to image.
    values = np.copy(image)
    bady = 0
    badx = 1
    if len(image.shape) == 3:
        bady = 1
        badx = 2
    elif len(image.shape) == 4:
        bady = 2
        badx = 3

    bad = np.where(values > limits)
    values[bad] = limits[bad[bady], bad[badx]]
    ncoeff = coeffs.shape[0]
    t = np.copy(coeffs[-1, :, :])
    for i in range(ncoeff-2, -1, -1):
        t = coeffs[i, :, :] + values*t
    return t


def nonLinDeriv(image, coeffs, limits):
    # First derivative of non-lin correction
    # function
    values = np.copy(image)
    bady = 0
    badx = 1
    if len(image.shape) == 3:
        bady = 1
        badx = 2
    elif len(image.shape) == 4:
        bady = 2
        badx = 3

    bad = np.where(values > limits)
    values[bad] = limits[bad[bady], bad[badx]]
    ncoeff = coeffs.shape[0]
    t = (ncoeff-1) * np.copy(coeffs[-1, :, :])
    for i in range(ncoeff-3, -1, -1):
        t = (i+1) * coeffs[i+1, :, :] + values * t
    return t



"""Unit tests for background generation

Authors
-------
    - Bryan Hilbert

Use
---
    Ensure you have pytest installed. Then, simply run pytest in any
    parent directory of mirage/tests/:
    >>> pytest
"""
import astropy.units as u
import datetime
from jwst_backgrounds import jbt
import numpy as np
import os
import pkg_resources
import pysiaf

from mirage.utils import backgrounds, file_io
from mirage.utils.constants import MEAN_GAIN_VALUES

package_path = pkg_resources.resource_filename('mirage', '')
CONFIG_DIR = os.path.join(package_path, 'config')


def test_day_of_year_background_spectrum():
    """Make sure the function returns a spectrum of wavelength and
    flux values.
    """
    ra = 57.2
    dec = -27.6
    observation_date = '2020-12-10'
    waves, signals = backgrounds.day_of_year_background_spectrum(ra, dec, observation_date)

    assert isinstance(waves, np.ndarray)
    assert isinstance(signals, np.ndarray)
    assert len(waves.shape) == 1
    assert len(signals.shape) == 1

    background = jbt.background(ra, dec, 4.)
    obsdate = datetime.datetime.strptime(observation_date, '%Y-%m-%d')
    obs_dayofyear = obsdate.timetuple().tm_yday

    assert obs_dayofyear in background.bkg_data['calendar']


def test_specific_day_of_year_background_spectrum():
    """Test this function using specific inputs and compare to the ETC
    output values"""
    sw_gain = MEAN_GAIN_VALUES['nircam']['swa']
    lw_gain = MEAN_GAIN_VALUES['nircam']['lwa']
    lw_etc = 2.26 / lw_gain  # 2.26 e/s/pix divided by gain 2.19 e/ADU, FOR LWA
    sw_etc = 0.20  / sw_gain  # 0.20 e/s/pix divided by gain 2.44 e/ADU for SWA

    # Use the NIRISS Focus Field
    ra = 85.22458
    dec = -69.5225
    obs_date = '2021-10-04'

    lw_filter_file = os.path.join(CONFIG_DIR, 'F444W_nircam_plus_ote_throughput_moda_sorted.txt')
    #lw_photflam = 7.7190e-22  # FLAM in cgs
    #lw_pivot = 4.3849  # microns
    lw_siaf = pysiaf.Siaf('nircam')['NRCA5_FULL']
    # Here: etc is 1.03, mirage is 0.84. This may be due to a bug in the ETC.

    sw_filter_file = os.path.join(CONFIG_DIR, 'F090W_nircam_plus_ote_throughput_moda_sorted.txt')
    #sw_photflam = 3.3895e-20  # FLAM in cgs
    #sw_pivot = 0.9034  # microns
    sw_siaf = pysiaf.Siaf('nircam')['NRCA2_FULL']

    waves, signals = backgrounds.day_of_year_background_spectrum(ra, dec, obs_date)

    sw_bkgd = backgrounds.calculate_background(ra, dec, sw_filter_file, True, sw_gain, sw_siaf,
                                               back_wave=waves, back_sig=signals)
    lw_bkgd = backgrounds.calculate_background(ra, dec, lw_filter_file, True, lw_gain, lw_siaf,
                                               back_wave=waves, back_sig=signals)

    assert np.isclose(sw_bkgd, sw_etc, atol=0, rtol=0.15)
    assert np.isclose(lw_bkgd, lw_etc, atol=0, rtol=0.25)

def test_find_low_med_high():
    """Test function for finding pre-defined low, medium, and high
    percentile values from an array"""
    array = np.arange(10) + 1
    values = backgrounds.find_low_med_high(array)
    assert values == [1., 5., 9.]


def test_find_percentile():
    """Test the find_percentile function"""
    array = np.arange(10) + 1
    perc = backgrounds.find_percentile(array, 0.6)
    assert perc == 6

    perc = backgrounds.find_percentile(array, 0.2)
    assert perc == 2


def test_low_med_high_background_spectrum():
    """Test the calculation of a background spectrum based on low, medium,
    high values"""
    detector = 'NRCA1'
    module = 'A'
    params = {'Telescope': {'ra': 57.2, 'dec': -27.6},
              'Readout': {'filter': 'F200W', 'pupil': 'CLEAR'},
              'simSignals': {'bkgdrate': 'low'},
              'Inst': {'instrument': 'NIRCAM'},
              'Reffiles': {'flux_cal': os.path.join(CONFIG_DIR, 'NIRCam_zeropoints.list')}
              }

    waves, signals = backgrounds.low_med_high_background_spectrum(params, detector, module)

    assert isinstance(waves, np.ndarray)
    assert isinstance(signals, np.ndarray)
    assert len(waves.shape) == 1
    assert len(signals.shape) == 1


def test_low_medium_high_background_value():
    """Test that the proper integrated background value is calculated for
    a given filter and level"""
    ra = 57.2
    dec = -27.6

    filt_waves = np.array([1., 2., 2.5, 2.6, 2.7, 2.8, 2.9, 3.0, 3.1, 4., 5.])
    filt_thru = np.array([0., 0., 0., 1., 1., 1., 1., 1., 0., 0., 0.])

    siaf = pysiaf.Siaf('nircam')['NRCA1_FULL']
    bkgd_high = backgrounds.low_medium_high_background_value(ra, dec, "high", filt_waves, filt_thru, siaf)
    bkgd_med = backgrounds.low_medium_high_background_value(ra, dec, "medium", filt_waves, filt_thru, siaf)
    bkgd_low = backgrounds.low_medium_high_background_value(ra, dec, "low", filt_waves, filt_thru, siaf)

    assert bkgd_high > bkgd_med
    assert bkgd_med > bkgd_low


def test_specific_low_medium_high_background_value():
    """Test specific cases of this function and compare to ETC outputs
    """
    # etc = {'low': 0.24, 'medium': 0.26, 'high': 0.27}  # MJy/sr from webform
    etc = {'low': 1.246, 'medium': 1.352, 'high': 1.402}  # e-/sec/pixel measured from output image

    # Use the NIRISS Focus Field
    ra = 85.22458
    dec = -69.5225

    siaf = pysiaf.Siaf('niriss')['NIS_CEN']
    filter_throughput_file = os.path.join(CONFIG_DIR, 'f150w_niriss_throughput1.txt')
    filter_waves, filter_thru = file_io.read_filter_throughput(filter_throughput_file)

    bkgd_high = backgrounds.low_medium_high_background_value(ra, dec, "high", filter_waves, filter_thru, siaf)
    bkgd_med = backgrounds.low_medium_high_background_value(ra, dec, "medium", filter_waves, filter_thru, siaf)
    bkgd_low = backgrounds.low_medium_high_background_value(ra, dec, "low", filter_waves, filter_thru, siaf)

    assert np.isclose(bkgd_high, etc['high'], atol=0., rtol=0.05)
    assert np.isclose(bkgd_med, etc['medium'], atol=0., rtol=0.05)
    assert np.isclose(bkgd_low, etc['low'], atol=0., rtol=0.05)


def test_niriss_background_scaling():
    """Test that the scaling factor for NIRISS is correctly calculated"""
    detector = 'NIS'
    module = 'N'
    params = {'Telescope': {'ra': 57.2, 'dec': -27.6},
              'Readout': {'filter': 'F090W', 'pupil': 'CLEAR'},
              'simSignals': {'bkgdrate': 'medium'},
              'Inst': {'instrument': 'NIRISS'},
              'Reffiles': {'flux_cal': os.path.join(CONFIG_DIR, 'niriss_zeropoints.list')}
              }

    # Medium-scaled background
    medium = backgrounds.niriss_background_scaling(params, detector, module)

    # Low-scaled background
    params['simSignals']['bkgdrate'] = 'low'
    low = backgrounds.niriss_background_scaling(params, detector, module)
    assert low < medium



#! /usr/bin/env python
# coding=UTF-8
# Client for launching simulations for the Besancon Galaxy Model web service and Gravpot web service.
#
#
# Modified to work with Mirage
#
#
# Version 1.0,
# Date: 2019 january 14th.
# Author: Raphael Melior
# Institut UTINAM, OSU THETA, France
#
# .Change log:
#   - 2019-01-19:
#        model URL turned from http to https
#   - 2019-01-21
#        compatible python3 (add parentesis of print functions, long() => int())

# http://docs.python-requests.org/en/latest/user/quickstart/
# https://docs.python.org/2/library/xml.etree.elementtree.html

#________________________
#Example of usage for gravpot web service:
#   How to launch a simulation :
#./galmod_client.py --url "https://gravpot.utinam.cnrs.fr/" --user username --create -p RA 155.6 -p pmRA 2.5 -p Dec -12.5 -p pmDec -5.6 -p d 1.1 -p RV 15 --run
#
#   To see the status of a job :
#./galmod_client.py --url "https://gravpot.utinam.cnrs.fr/" --user username --job jobnumber
#
# see below (in paragraphe "main") various possible options and actions)
#________________________

import argparse
import datetime
import dateutil.parser
import requests # installs with : pip install requests
import xml.etree.ElementTree as ElementTree
import os
import getpass


XML_NS = {'uws':'http://www.ivoa.net/xml/UWS/v1.0', 'xlink':'http://www.w3.org/1999/xlink'}

class UWS:

    def __init__(self, url, auth):
        self._url = url
        self._auth = auth

    def getJobsList(self):
        r = requests.get(self._url+"/jobs", auth=self._auth)
        return self._getJobInfosFromReqResult(r, "get list of jobs")

    def createJob(self):
        r = requests.post(self._url+"/jobs", auth=self._auth)
        return self._getJobInfosFromReqResult(r, "creating job")

    def getJobInfos(self, idJob):
        r = requests.get(self._url+"/jobs/"+str(idJob), auth=self._auth)
        return self._getJobInfosFromReqResult(r, "get details of a job")

    def deleteJob(self, idJob):
        r = requests.delete(self._url+"/jobs/"+str(idJob), auth=self._auth)
        return self._getJobInfosFromReqResult(r, "deleting a job")

    def runJob(self, idJob):
        r = requests.post(self._url+"/jobs/"+str(idJob)+'/phase', auth=self._auth, data={'PHASE':'RUN'})
        return self._getJobInfosFromReqResult(r, "starting a job")

    def abortJob(self, idJob):
        r = requests.post(self._url+"/jobs/"+str(idJob)+'/phase', auth=self._auth, data={'PHASE':'ABORT'})
        return self._getJobInfosFromReqResult(r, "aborting a job")

    def maxDurationJob(self, idJob, seconds):
        r = requests.post(self._url+"/jobs/"+str(idJob)+'/executionduration', auth=self._auth, data={'EXECUTIONDURATION':seconds})
        return self._getJobInfosFromReqResult(r, "setting execution duration of a job")

    def timeDestructJob(self, idJob, timeStr):
        r = requests.post(self._url+"/jobs/"+str(idJob)+'/destruction', auth=self._auth, data={'DESTRUCTION':timeStr})
        return self._getJobInfosFromReqResult(r, "setting execution duration of a job")

    def setParams(self, idJob, params):
        #if len(params) == 1:   # TODO
        #   r = requests.put(self._url+"/jobs/"+str(idJob)+'/parameters/'+params[0][0], data=params[0][1])
        #   print(r.text)
        #   return getJobInfos(idJob)#self._getJobInfosFromReqResult(r, "setting parameters of a job")
        #elif len(params) > 1:
            data={}
            for p in params:
                data[p[0]] = p[1]
            r = requests.post(self._url+"/jobs/"+str(idJob)+'/parameters', auth=self._auth, data=data)
            return self._getJobInfosFromReqResult(r, "setting parameters of a job")

    def _getJobInfosFromReqResult(self, r, phase):
        if (len(r.history) > 0):
            print("SERVER RESPONSE : "+r.history[0].text)

        if (r.status_code == 200):
            #et = ElementTree
            #ElementTree.register_namespace('uws', 'http://www.ivoa.net/xml/UWS/v1.0')
            for ns in XML_NS:
                ElementTree.register_namespace(ns, XML_NS[ns])

            try:
                return ElementTree.fromstring(r.text)
            except Exception as e:
                print("Invalid response :")
                print(r.text)
                raise e
        else:
            raise JobError(r, phase)




# Convertion into text

def xmlTextIfFound(xmlElem):
    if (xmlElem == None) or (xmlElem.text == None):
        return ''
    else:
        return xmlElem.text

def xmlDateIfFound(xmlElem):
    if (xmlElem == None) or (xmlElem.text == None):
        return '-'
    else:
        d = dateutil.parser.parse(xmlElem.text).astimezone(dateutil.tz.tzutc())
        return d.strftime("%Y-%m-%d %H:%M")


# Exceptions

class JobError(Exception):
    def __init__(self, http, when):
        self.status = http.status_code
        self.explain = http.text
        self.when = when
    def __str__(self):
        return "Error when "+self.when+" : "+str(self.status)+" "+self.explain



# Main

def run(ra, dec, ra_width, dec_width, user, password, url='https://model.obs-besancon.fr/ws', create=None, job=None, ):
    """
    """
    params = [('RA', ra), ('Dec', dec), ('pmRA', ra_width), ('pmDec', dec_width)]


    uws = UWS(url, (user, passwordd))
    try:
        print("Creating job")
        job = uws.createJob()
    except JobError as e:
        print(e)
        exit(2)

    idJob = int(job.find('uws:jobId', XML_NS).text)

    job = uws.setParams(idJob, args.param)

    job = uws.runJob(idJob)

    print("id                :"+str(idJob))
    print("owner             :"+xmlTextIfFound(job.find('uws:ownerId', XML_NS)))
    print("phase             :"+xmlTextIfFound(job.find('uws:phase', XML_NS)))
    print("startTime         :"+xmlTextIfFound(job.find('uws:startTime', XML_NS)))
    print("endTime           :"+xmlTextIfFound(job.find('uws:endTime', XML_NS)))
    print("destruction       :"+xmlTextIfFound(job.find('uws:destruction', XML_NS)))
    print("executionDuration :"+xmlTextIfFound(job.find('uws:executionDuration', XML_NS)))

    parameters = job.find('uws:parameters', XML_NS)
    if parameters is not None:
        print('\nParameters ........')
        for param in parameters.findall('uws:parameter', XML_NS):
            print('{:<18}:{}'.format(param.get('id'), xmlTextIfFound(param)))

    errorSummary = job.find('uws:errorSummary', XML_NS)
    if errorSummary is not None:
        print('\nErrors ............')
        for err in errorSummary.findall('uws:message', XML_NS):
            print('error             :'+xmlTextIfFound(err))

    results = job.find('uws:results', XML_NS)
    if results is not None:
        print('\nResults ...........')
        for res in results.findall('uws:result', XML_NS):
            print('{:<18}:{}'.format(res.get('id'), res.get('{http://www.w3.org/1999/xlink}href')))






if __name__=="__main__":
    parser = argparse.ArgumentParser(description='Makes requests to web services, either the Besancon Model of the Galaxy or Gravpot web service', add_help=False)

    serverGrp = parser.add_argument_group('Server options', 'Options for server connection.')
    serverGrp.add_argument('--url', help='URL of the server.', default='https://model.obs-besancon.fr/ws')
    serverGrp.add_argument('--user', help='Username used to connect to server, by default it is the system username ('+os.getlogin()+').', default=os.getlogin())
    serverGrp.add_argument('--pass', dest='passwd', help='Password used to connect to server, by default the password is prompted. Usage of this option is discouraged because it will be visible, for example, with "ps" command and written to history.')

    jobGrp = parser.add_argument_group('Job selection', 'Select the job you want to manipulate. To create a new one type "--create". If this option is not present, it returns your list of jobs and exit ignoring options in "Actions".').add_mutually_exclusive_group()
    jobGrp.add_argument('-c', '--create', help='Create a new job.', action='store_true')
    jobGrp.add_argument('-j', '--job', type=int, help='The job affected by this request.')

    actionGrp = parser.add_argument_group('Actions', 'Configure and make actions on the selected job.')
    actionGrp.add_argument('-p', '--param', help='Set parameter named PARAMNAME with VALUE.', action='append', nargs=2, metavar=('PARAMNAME', 'VALUE'))
    actionGrp.add_argument('--execdur', help='Set maximum execution duration of a job in seconds (this setting can be overridden by server\'s configuration).', type=int)
    actionGrp.add_argument('--tdest', help='Set destruction time of a job (date+time in ISO8601 format, this setting can be overridden by server\'s configuration).')
    actionGrp.add_argument('--run', help='Send job for computation.', action='store_true')
    actionGrp.add_argument('--abort', help='Abort the job.', action='store_true')
    actionGrp.add_argument('--delete', help='Delete the job.', action='store_true')

    miscGrp = parser.add_argument_group('Miscellaneous')
    miscGrp.add_argument('-h', '--help', help='Show this help message and exit.', action='store_true')

    args = parser.parse_args()

    if (args.help):
        parser.print_help()
        exit(0)

    if (args.passwd != None):
        passwd = args.passwd
    else:
        passwd = getpass.getpass('Password for '+args.user+': ')

    #print(args.user+' '+passwd)
    uws = UWS(args.url, (args.user, passwd))

    #print(args) #args.output
    job = None
    if (args.create):
        try:
            print("Creating job")
            job = uws.createJob()
        except JobError as e:
            print(e)
            exit(2)

    elif (args.job != None):
        try:
            job = uws.getJobInfos(args.job)
        except JobError as e:
            print(e)
            exit(2)

    else:
        print(" JobID        phase       startTime          endTime            destruction    ")
        print("----------   ---------   ----------------   ----------------   ----------------")
        #print("0123456789   COMPLETED   2009-05-19 17:12   2009-05-19 17:15   2009-06-06 17:12")
        jobs = uws.getJobsList()
        for job in jobs.findall('uws:job', XML_NS):
            jobId = xmlTextIfFound(job.find('uws:jobId', XML_NS))
            phase = xmlTextIfFound(job.find('uws:phase', XML_NS))
            startTime = xmlDateIfFound(job.find('uws:startTime', XML_NS))
            endTime = xmlDateIfFound(job.find('uws:endTime', XML_NS))
            destruction = xmlDateIfFound(job.find('uws:destruction', XML_NS))
            print("%10d   %9s   %16s   %16s   %16s" % (int(jobId), phase, startTime, endTime, destruction))
        exit(0)


    idJob = int(job.find('uws:jobId', XML_NS).text)

    if args.param:
        print(args.param)
        job = uws.setParams(idJob, args.param)

    if args.execdur:
        job = uws.maxDurationJob(idJob, args.execdur)

    if args.tdest:
        job = uws.timeDestructJob(idJob, args.tdest)

    if args.run:
        job = uws.runJob(idJob)
    elif args.abort:
        job = uws.abortJob(idJob)

    print("id                :"+str(idJob))
    print("owner             :"+xmlTextIfFound(job.find('uws:ownerId', XML_NS)))
    print("phase             :"+xmlTextIfFound(job.find('uws:phase', XML_NS)))
    print("startTime         :"+xmlTextIfFound(job.find('uws:startTime', XML_NS)))
    print("endTime           :"+xmlTextIfFound(job.find('uws:endTime', XML_NS)))
    print("destruction       :"+xmlTextIfFound(job.find('uws:destruction', XML_NS)))
    print("executionDuration :"+xmlTextIfFound(job.find('uws:executionDuration', XML_NS)))

    parameters = job.find('uws:parameters', XML_NS)
    if parameters is not None:
        print('\nParameters ........')
        for param in parameters.findall('uws:parameter', XML_NS):
            print('{:<18}:{}'.format(param.get('id'), xmlTextIfFound(param)))

    errorSummary = job.find('uws:errorSummary', XML_NS)
    if errorSummary is not None:
        print('\nErrors ............')
        for err in errorSummary.findall('uws:message', XML_NS):
            print('error             :'+xmlTextIfFound(err))

    results = job.find('uws:results', XML_NS)
    if results is not None:
        print('\nResults ...........')
        for res in results.findall('uws:result', XML_NS):
            print('{:<18}:{}'.format(res.get('id'), res.get('{http://www.w3.org/1999/xlink}href')))

    if args.delete:
        uws.deleteJob(idJob)



'''Simulate a commissioning PSF using WebbPSF.

*** NOT CURRENTLY IN USE ****
'''

import webbpsf


def create_commissioning_psf(params):
    '''Placeholder function for ultimate WebbPSF query.

    Arguments
    ---------
    params
        Set of parameters that will be used to make a custom PSF with
        WebbPSF (i.e. detector, filter, countrate, sub-pixel position,
        APT #, observation/visit #, etc...)
    '''
	return None


#! /usr/bin env python
"""This module contains functions for calculating background signals using
jwst_backgrounds
"""
import astropy.units as u
import datetime
import numpy as np

from jwst_backgrounds import jbt

from mirage.utils.constants import PRIMARY_MIRROR_AREA, PLANCK
from mirage.utils.file_io import read_filter_throughput
from mirage.utils.flux_cal import fluxcal_info

# Percentiles corresponding to "low", "medium", and "high" as used in
# the ETC
LOW = 0.1
MEDIUM = 0.5
HIGH = 0.9


def calculate_background(ra, dec, filter_file, use_dateobs, gain_value,
                         siaf_instance, back_wave=None, back_sig=None, level='medium'):
    """Use the JWST background calculator to come up with an appropriate background
    level for the observation.

    Parameters
    ----------
    ra : float
        RA in degrees

    dec : float
        Dec in degrees

    filter_file : str
        Name of ascii file containing filter throughput curve

    use_dateobs : bool
        Use the observation date to find the background value

    photflam : float
        Conversion factor from FLAM to ADU/sec

    pivot : float
        Pivot wavelength for the filter correspondig to ``filter_file``

    siaf_instance : pysiaf.Siaf
        Siaf instance for the instrument/aperture to be used

    back_wave : numpy.ndarray
        1D array of wavelength values for the background spectrum.
        These are only used in the case where ``use_dateobs`` is True

    back_sig : numpy.ndarray
        1D array of signal values for the background spectrum.
        These are only used in the case where ``use_dateobs`` is True

    level : str
        'low', 'medium', or 'high'

    Returns
    -------
    bval.value : float
        Background value in units of ADU/sec/pixel
    """
    from jwst_backgrounds import jbt
    from astropy import units as u
    from astropy.units.equivalencies import si, cgs

    # Read in filter throughput file
    filt_wav, filt_thru = read_filter_throughput(filter_file)

    # If the user wants a background signal from a particular day,
    # then extract that array here
    if use_dateobs:
        # Interpolate background to match filter wavelength grid
        bkgd_interp = np.interp(filt_wav, back_wave, back_sig)

        # Combine
        filt_bkgd = bkgd_interp * filt_thru

        pixelarea = siaf_instance.XSciScale * u.arcsec * siaf_instance.YSciScale * u.arcsec
        photon_total = PRIMARY_MIRROR_AREA * (filt_bkgd * u.MJy / u.sr) * (1. / PLANCK) * 1.e-20 * pixelarea.to(u.sr) / (filt_wav * u.micron)
        bval = np.trapz(photon_total, x=filt_wav)
        bval = bval.value

    else:
        # If the user has requested background in terms of low/medium/high,
        # then we need to examine all the background arrays.
        # Loop over each day (in the background)
        # info, convolve the background curve with the filter
        # throughput curve, and then integrate. THEN, we can
        # calculate the low/medium/high values.
        bval = low_medium_high_background_value(ra, dec, level, filt_wav, filt_thru, siaf_instance)

    # Convert the background signal from e-/sec/pixel to ADU/sec/pixel
    bval /= gain_value
    return bval


def day_of_year_background_spectrum(ra, dec, observation_date):
    """Call jwst_backgrounds in order to produce an estimate of background
    versus wavelength for a particular pointing on a particular day of
    year.

    Parameters
    ----------
    ra : float
        Right Ascention of pointing

    dec : float
        Declination of pointing

    observation_date : str
        Requested day of year of the observation i.e. '2021-10-31'

    Returns
    -------
    background_waves : numpy.ndarray
        1D array with the wavelength values in microns associated with
        ``background_signals``

    background_sigmals : numpy.ndarray
        1D array containing background values in MJy/str
    """
    # Generate background spectra for all days
    background = jbt.background(ra, dec, 4.)

    # Confirm that the target is observable of the requested day of year
    obsdate = datetime.datetime.strptime(observation_date, '%Y-%m-%d')
    obs_dayofyear = obsdate.timetuple().tm_yday
    if obs_dayofyear not in background.bkg_data['calendar']:
        raise ValueError(("ERROR: The requested RA, Dec is not observable on {}. Either "
                          "specify a different day, or set simSignals:use_dateobs_for_background "
                          "to False.".format(observation_date)))

    # Extraxct the spectrum for the requested day
    match = obs_dayofyear == background.bkg_data['calendar']
    background_waves = background.bkg_data['wave_array']
    background_signals = background.bkg_data['total_bg'][match, :][0]

    return background_waves, background_signals


def find_low_med_high(array):
    """Given an array of values, find the value corresponding to the
    Nth percentile.

    Parameters
    ----------
    array : numpy.ndarray
        1D array of values

    Returns
    -------
    levels : list
        [low, medium, high] percentile values of array
    """
    levels = []
    for value in [LOW, MEDIUM, HIGH]:
        levels.append(find_percentile(array, value))
    return levels


def find_percentile(array, percentile):
    """Find the value corresponding to the ``percentile``
    percentile of ``array``.

    Parameters
    ----------
    array : numpy.ndarray
        Array of values to be searched

    percentile : float
        Percentile to search for. For example, to find the 50th percentile
        value, use 0.5
    """
    x = np.sort(array)
    y = np.arange(1, len(x) + 1) / len(x)
    value = np.interp(percentile, y, x)
    return value


def low_med_high_background_spectrum(param_dict, detector, module):
    """Call jwst_backgrounds in order to produce an estimate of background
    versus wavelength for a particular pointing that corresponds to one of
    "low", "medium", or "high", as with the ETC.

    Parameters
    ----------
    param_dict : dict
        Dictionary of observation information from an input yaml file

    detector : str
        Name of detector, e.g. "NRCA1"

    module : str
        Name of module, e.g. "A"

    Returns
    -------
    background_waves : numpy.ndarray
        1D array with the wavelength values in microns associated with
        ``background_signals``

    background_spec : numpy.ndarray
        1D array containing background values in MJy/str

    """
    # Generate background spectra for all days
    background = jbt.background(param_dict['Telescope']['ra'], param_dict['Telescope']['dec'], 4.)

    # Determine which optical element wheel contains the filter we want
    # to use for background estimation
    if param_dict['Readout']['pupil'][0].upper() == 'F':
        usefilt = 'pupil'
    else:
        usefilt = 'filter'

    # Get basic flux calibration information
    vegazp, photflam, photfnu, pivot_wavelength = fluxcal_info(param_dict, usefilt, detector, module)

    # Extract the spectrum value across all days at the pivot wavelength
    wave_diff = np.abs(background.bkg_data['wave_array'] - pivot_wavelength)
    bkgd_wave = np.where(wave_diff == np.min(wave_diff))[0][0]
    bkgd_at_pivot = background.bkg_data['total_bg'][:, bkgd_wave]

    # Now sort and determine the low/medium/high levels
    low, medium, high = find_low_med_high(bkgd_at_pivot)

    # Find the value based on the level in the yaml file
    background_level = param_dict['simSignals']['bkgdrate'].lower()
    if background_level == "low":
        level_value = low
    elif background_level == "medium":
        level_value = medium
    elif background_level == "high":
        level_value = high
    else:
        raise ValueError(("ERROR: Unrecognized background value: {}. Must be low, mediumn, or high"
                          .format(param_dict['simSignals']['bkgdrate'])))

    # Find the day with the background at the pivot wavelength that
    # is closest to the value associated with the requested level
    diff = np.abs(bkgd_at_pivot - level_value)
    mindiff = np.where(diff == np.min(diff))[0][0]
    background_spec = background.bkg_data['total_bg'][mindiff, :]

    return background.bkg_data['wave_array'], background_spec


def low_medium_high_background_value(ra, dec, background_level, filter_waves, filter_throughput, siaf_info):
    """Calculate the integrated background flux density for a given filter,
    using the filter's throughput curve and the user-input background level
    (e.g. "medium")

    Parameters
    ----------
    ra : float
        Right ascention of the pointing. Units are degrees

    dec : float
        Declineation of the pointing. Units are degrees

    background_level : str
        "low", "medium", or "high", just as with the ETC

    filter_waves : numpy.ndarray
        1d array of wavelengths in microns to be used along with
        ``filter_throughput``

    filter_throughput : numpy.ndarray
        1d array of filter throughput values to convolve with the background
        spectrum. Normalized units. 1.0 = 100% transmission.

    siaf_info : pysiaf.Siaf
        Siaf information for the detector/aperture in use

    Returns
    -------
    value : float
        Background value corresponding to ``background_level``, integrated
        over the filter bandpass. Background units are e-/sec/pixel.
    """
    # Get background information
    bg = jbt.background(ra, dec, 4.)
    back_wave = bg.bkg_data['wave_array']
    bsigs = np.zeros(len(bg.bkg_data['total_bg'][:, 0]))
    for i in range(len(bg.bkg_data['total_bg'][:, 0])):
        back_sig = bg.bkg_data['total_bg'][i, :]

        # Interpolate background to match filter wavelength grid
        bkgd_interp = np.interp(filter_waves, back_wave, back_sig)

        # Combine
        filt_bkgd = bkgd_interp * filter_throughput

        # Convert from MJy/sr to e-/sec/pixel
        pixelarea = siaf_info.XSciScale * u.arcsec * siaf_info.YSciScale * u.arcsec
        photon_total = PRIMARY_MIRROR_AREA * (filt_bkgd * u.MJy / u.sr) * (1. / PLANCK) * 1.e-20 * pixelarea.to(u.sr) / (filter_waves * u.micron)
        bsigs[i] = np.trapz(photon_total.value, x=filter_waves)

    # Now sort and determine the low/medium/high levels
    low, medium, high = find_low_med_high(bsigs)

    # Find the value based on the level in the yaml file
    background_level = background_level.lower()
    if background_level == "low":
        value = low
    elif background_level == "medium":
        value = medium
    elif background_level == "high":
        value = high
    else:
        raise ValueError(("ERROR: Unrecognized background value: {}. Must be low, mediumn, or high"
                          .format(background_level)))
    return value


def nircam_background_spectrum(parameters, detector, module):
    """Generate a background spectrum by calling jwst_backgrounds and
    returning wavelengths and flux density values based on observation
    date or low/medium/high

    Parameters
    ----------
    parameters : dict
        Nested dictionary containing parameters pertaining to background
        generation. Designed around dictionary from reading in a Mirage
        input yaml file

    detector : str
        Detector name (e.g. 'NRCA1')

    module : str
        Module name (e.g. 'A')

    Returns
    -------
    waves : numpy.ndarray
        1D array of wavelength values (in microns)

    fluxes : numpy.ndarray
        1D array of flux density values
    """
    if parameters['simSignals']['use_dateobs_for_background']:
        print("Generating background spectrum for observation date: {}"
              .format(parameters['Output']['date_obs']))
        waves, fluxes = day_of_year_background_spectrum(parameters['Telescope']['ra'],
                                                        parameters['Telescope']['dec'],
                                                        parameters['Output']['date_obs'])
    else:
        if isinstance(parameters['simSignals']['bkgdrate'], str):
            if parameters['simSignals']['bkgdrate'].lower() in ['low', 'medium', 'high']:
                print("Generating background spectrum based on requested level of: {}"
                      .format(parameters['simSignals']['bkgdrate']))
                waves, fluxes = low_med_high_background_spectrum(parameters, detector, module)
            else:
                raise ValueError("ERROR: Unrecognized background rate. Must be one of 'low', 'medium', 'high'")
        else:
            raise ValueError(("ERROR: WFSS background rates must be one of 'low', 'medium', 'high', "
                              "or use_dateobs_for_background must be True "))
    return waves, fluxes


def niriss_background_scaling(param_dict, detector, module):
    """Determine the scaling factor needed to translate the pre-existing
    NIRISS WFSS background image to the requested signal level, which is
    one of "low", "medium", or "high", as with the ETC.

    Parameters
    ----------
    param_dict : dict
        Dictionary of observation information from an input yaml file

    detector : str
        Name of detector, e.g. "NRCA1"

    module : str
        Name of module, e.g. "A"

    Returns
    -------
    ratio : float
        Ratio of the background value at the pivot wavelength that
        corresponds to the requested level, ratioed to that for the
        "medium" level, which is what was used to create the NIRISS
        WFSS background images.
    """
    # Generate background spectra for all days
    background = jbt.background(param_dict['Telescope']['ra'], param_dict['Telescope']['dec'], 4.)

    # Determine which optical element wheel contains the filter we want
    # to use for background estimation
    if param_dict['Readout']['pupil'][0].upper() == 'F':
        usefilt = 'pupil'
    else:
        usefilt = 'filter'

    # Get basic flux calibration information
    vegazp, photflam, photfnu, pivot_wavelength = fluxcal_info(param_dict, usefilt, detector, module)

    # Extract the spectrum value across all days at the pivot wavelength
    wave_diff = np.abs(background.bkg_data['wave_array'] - pivot_wavelength)
    bkgd_wave = np.where(wave_diff == np.min(wave_diff))[0][0]
    bkgd_at_pivot = background.bkg_data['total_bg'][:, bkgd_wave]

    # Now sort and determine the low/medium/high levels
    low, medium, high = find_low_med_high(bkgd_at_pivot)

    # Find the value based on the level in the yaml file
    background_level = param_dict['simSignals']['bkgdrate'].lower()
    if background_level == "low":
        level_value = low
    elif background_level == "medium":
        level_value = medium
    elif background_level == "high":
        level_value = high
    else:
        raise ValueError(("ERROR: Unrecognized background value: {}. Must be low, medium, or high"
                          .format(params_dict['simSignals']['bkgdrate'])))
    return level_value



#! /usr/bin/env python

"""
This module contains tools for adding a source with time-varying signal
to a 4-dimensional seed image
"""
import copy
import numpy as np
import h5py
from scipy.integrate import romb

from mirage.catalogs.hdf5_catalog import open_tso


def add_tso_sources(seed_image, seed_segmentation_map, psf_seeds, segmentation_maps, lightcurves, frametime,
                    total_frames, exposure_total_frames, frames_per_integration, number_of_ints, resets_bet_ints,
                    starting_time=0, starting_frame=0, samples_per_frametime=5):
    """inputs are lists, so that there can be multiple sources, although this
    situation is extremely unlikely

    Parameters
    ----------
    seed_image : numpy.ndarray
        2D seed image in units of counts per second

    seed_segmentation_map : numpy.ndarray
        2D segmentation map associated with ``seed_image``

    psf_seeds : list
        List of 2D seed images, each containing only the signal
        (counts/sec) for a single TSO target.

    segmentation_maps : list
        List of 2D segmentation maps associated with ``psf_seeds``

    lightcurves : list
        List of dictionaries containing lightcurve data. (i.e. output of
        ``read_lightcurve``

    frametime : float
        Exposure time associated with a single frame

    total_frames : int
        Total number of frames in the observation (this includes
        resets between integrations)

    exposure_total_frames : int
        Number of frames in the exposure (excluding resets between
        integrations)

    frames_per_integration : int
        Number of frames per integration of the exposure

    number_of_ints : int
        Number of integrations in the exposure to be simulated

    resets_bet_ints : int
        Number of reset frames between integrations

    samples_per_frametime : int
        Number of points per ``frametime`` to interpolate the lightcurve
        to. For each frame, the lightcurve will be integrated across
        these points using Romberg Integration. This means that
        samples_per_frametime must be 2^k + 1 for some integer k.

    Returns
    -------
    final_seed : numpy.ndarray
        4D array containing the 2D seed image for each frame of each
        integration in the exposure

    final_seed_segmentation_map : numpy.ndarray
        2D array containing the segmentation map
    """
    yd, xd = seed_image.shape
    total_exposure_time = exposure_total_frames * frametime

    # Make sure that samples_per_frametime has a value that will allow
    # Romberg integration
    k_remainder, k = np.modf(np.log2(samples_per_frametime - 1))
    if k_remainder > 0.01:
        raise ValueError("samples_per_frametime must be 2^k + 1 for some integer k.")

    # Put seed image in units of counts per frame
    # seed_image *= frametime
    seed_image_per_frame = seed_image * frametime

    # Frame seed contains only the signal for that particular frame,
    # rather than the cumulative signal since the beginning of the
    # integration
    frame_seed = np.zeros((total_frames, yd, xd))
    final_seed_segmentation_map = np.zeros((yd, xd))
    # Loop over TSO objects
    for source_number, (psf, seg_map, lightcurve) in enumerate(zip(psf_seeds, segmentation_maps, lightcurves)):
        # Check that the provided lightcurve is long enough to cover the
        # exposure time of the observation. If not...extend the lightcurve
        # with values of 1.0 until is is long enough. If the lightcurve
        # starts at some non-zero time, then extend the lightcurve with
        # 1.0's back to time=0.
        lightcurve = check_lightcurve_time(lightcurve, total_exposure_time, frametime)

        # Scale the TSO source's seed image contribution to be for one
        # frametime rather than 1 second
        ft_psf = psf * frametime

        # Interpolate the lightcurve to prepare for integration
        interp_lightcurve = interpolate_lightcurve(copy.deepcopy(lightcurve), samples_per_frametime, frametime)
        dx = frametime / (samples_per_frametime - 1)

        # Integrate the lightcurve for each frame
        print('\nIntegrating lightcurve signal for each frame ')
        for frame_number in np.arange(total_frames) + starting_frame:
            frame_index = frame_number - starting_frame
            #print("Loop 1, frame number and index: ", frame_number, frame_index)
            min_index = frame_number * (samples_per_frametime - 1)
            indexes = np.arange(min_index, min_index + samples_per_frametime)

            # Normalize the integrated signal by the frametime, as that
            # is the integral of a flat line at 1.0 over one frametime
            relative_signal = romb(interp_lightcurve['fluxes'].value[indexes], dx) / frametime
            frame_psf = ft_psf * relative_signal
            tmpy, tmpx = psf.shape
            frame_seed[frame_index, :, :] += frame_psf

        # Add the TSO target to the segmentation map
        final_seed_segmentation_map = update_segmentation_map(seed_segmentation_map, seg_map.segmap)

    # Translate the frame-by-frame seed into the final, cumulative seed
    # image. Rearrange into integrations, resetting the signal for each
    # new integration.
    print('Translate the frame-by-frame transit seed into the final, cumulative seed image.')
    integration_starts = np.arange(number_of_ints) * (frames_per_integration + resets_bet_ints) + starting_frame
    reset_frames = integration_starts[1:] - 1

    if total_frames-len(reset_frames) > frames_per_integration:
        dimension = frames_per_integration
    else:
        dimension = total_frames-len(reset_frames)
    final_seed = np.zeros((number_of_ints, dimension, yd, xd))

    for frame in np.arange(total_frames) + starting_frame:
        int_number = np.where(frame >= integration_starts)[0][-1]
        rel_frame = frame - integration_starts[int_number]

        if frame in integration_starts:
            final_seed[int_number, 0, :, :] = copy.deepcopy(frame_seed[frame-starting_frame, :, :]) + seed_image_per_frame
        elif frame not in reset_frames:
            final_seed[int_number, rel_frame, :, :] = final_seed[int_number, rel_frame-1, :, :] + \
                frame_seed[frame-starting_frame, :, :] + seed_image_per_frame

    return final_seed, final_seed_segmentation_map


def check_lightcurve_time(light_curve, exposure_time, frame_time):
    """Check to be sure the provided lightcurve is long enough to cover
    the supplied total exposure time. If not, lengthen at the beginning
    or end such that it does. Times will only be added to the beginning
    if the first time entry in the lightcurve is > 0. Lightcurves where
    the initial time entry is < 0 will have all times < 0 chopped. This
    will allow the user to simulate lightcurves where the exposure starts
    somewhere in the middle of the lightcurve.

    Parameters
    ----------
    light_curve : dict
        Dictionary of lightcurve. "fluxes" and "times" keys contain
        arrays of those values

    exposure_time : float
        Total exposure time for the full exposure being simulated
        (in seconds)

    frame_time : float
        Exposure time of a single frame of the observation

    Returns
    -------
    light_curve : dict
        Potentially modified with added or removed elements
    """
    times = copy.deepcopy(light_curve["times"].value)
    fluxes = copy.deepcopy(light_curve["fluxes"].value)
    time_units = light_curve["times"].unit
    flux_units = light_curve["fluxes"].unit
    adjusted = False

    # Remove elements where time < 0.
    if np.min(times) < 0.:
        positive_times = times >= 0.
        times = times[positive_times]
        fluxes = fluxes[positive_times]
        adjusted = True

    # If the times begin at values significantly > 0,
    # then add entries to bring the start back to time = 0
    if np.min(times) > 0.:
        print(("Lightcurve time values do not start at zero. Prepending an entry with time=0 "
               "and flux = 1."))
        times = np.insert(times, 0, 0.)
        fluxes = np.insert(fluxes, 0, 1.)
        adjusted = True

    # If the ending time is less than the exposure's total
    # observation time, then add entries with flux=1
    if np.max(times) < exposure_time:
        print(("Lightcurve time values extend only to {} seconds. This is not long enough "
               "to cover the entire exposure time of {} seconds. Extending to cover the full "
               "exposure time with flux = 1.".format(np.max(times), exposure_time)))
        times = np.append(times, exposure_time + 5 * frame_time)
        fluxes = np.append(fluxes, 1.)
        adjusted = True

    if adjusted:
        light_curve["times"] = times * time_units
        light_curve["fluxes"] = fluxes * flux_units

    return light_curve


def interpolate_lightcurve(light_curve, samples_per_frame_time, frame_time):
    """Given a lightcurve with arbitrary sampling times, interpolate
    such that it has ``samples_per_frame_time`` samples per frametime.
    In order to use Romberg, we need to integrate over 2^k + 1 samples each time.

    Parameters
    ----------
    light_curve : dict
        Dictionary containing light curve. Times are in 'times' keyword,
        and fluxes are in 'fluxes' keyword. Each should be an array with
        associated unit.

    samples_per_frame_time : int
        Number of times to sample the lightcurve per frame exposure time

    frame_time : float
        Exposure time of one frame, in seconds.

    Returns
    -------
    light_curve : dict
        Modified dictionary containing the resampled lightcurve
    """
    time_units = light_curve['times'].unit
    flux_units = light_curve['fluxes'].unit
    divisor = samples_per_frame_time - 1.
    points = np.arange(light_curve['times'][0].value, light_curve['times'][-1].value, frame_time/divisor)
    light_curve["fluxes"] = np.interp(points, light_curve['times'].value, light_curve['fluxes'].value) * flux_units
    light_curve["times"] = points * time_units
    return light_curve


def read_lightcurve(filename, index_value):
    """Read in hdf5 file containing lightcurve data and return the
    lightcurve in the given index.

    Parameters
    ----------
    filename : str
        Name of hdf5 file containing lightcurves

    index_value : int
        Dataset value within ``filename`` continaing the lightcurve of
        interest

    Returns
    -------
    dataset : dict
        Dictionary containing lightcurve data for the object in dataset
        number ``index_value``.
    """
    file_contents = open_tso(filename)

    try:
        dataset = file_contents[index_value]
    except KeyError as e:
        print(e)
    return dataset


def update_segmentation_map(segmap, object_map):
    """Add the segmentation map for an individual object to that for the
    seed image. Note that multiple objects cannot be recorded in a single
    pixel. In this case, the TSO object of interest will be added to the
    segmentation map after the background objects, so it will overwrite
    any pixels that are affected by background objects.

    Parameters
    ----------
    segmap : numpy.ndarray
        2D segmenatation map of the scene

    object_map : numpy.ndarray
        2D segmentation map of object to add to the scene

    Returns
    -------
    segmap : numpy.ndarray
        Updated 2D segmentation map of the scene
    """
    obj_pix = object_map != 0
    segmap[obj_pix] = object_map[obj_pix]
    return segmap



"""System test of mirage/NIRISS for regular and NRM imaging.

Authors
-------
    - Kevin Volk

Use
---
    >>> pytest -s test_niriss_imaging_modes.py

Description of the test:
------------------------

The test runs mirage for the same NIRISs imaging scene in regular imaging and
in the NRM imaging mode.  There is only one star in the scene of the same
magnitude in both instances.  The code reads the two pointsource output list
files to get the count rates for the two cases and verifies that the proper
scaling has been used.  The count rate ratio should be exactly 0.15/0.84
between the NRM case and the regular imaging case.  Due to limitations on the
precision of the output format, the values differ by a fraction amount of
about 1.e-08 in my test.  Here the threhsold for agreement is a deviaiton of
less than 1.e-06.

"""

import os
import pytest
import numpy

from mirage import imaging_simulator

os.environ['TEST_DATA'] = os.path.join(os.path.dirname(__file__), 'test_data/NIRISS')

# Determine if tests are being run on Travis
ON_TRAVIS = 'travis' in os.path.expanduser('~')


@pytest.mark.skipif(ON_TRAVIS,
                   reason="Cannot access mirage data in the central storage directory from Travis CI.")
def test_niriss_imaging():
    nis = imaging_simulator.ImgSim(offline=True)
    nis.paramfile = os.path.join(os.path.dirname(__file__), 'test_data/NIRISS/niriss_imaging_test.yaml')
    nis.create()
    nis.paramfile = os.path.join(os.path.dirname(__file__), 'test_data/NIRISS/niriss_nrm_test.yaml')
    nis.create()
    value1 = numpy.loadtxt('V88888024002P000000000112o_NIS_F480M_uncal_pointsources.list',usecols=[8,])
    value2 = numpy.loadtxt('V88888024002P000000000112o_NIS_NRM_F480M_uncal_pointsources.list',usecols=[8,])
    fluxratio = value2 / value1
    targetratio = 0.15 / 0.84
    deviation = abs(fluxratio/targetratio - 1.)
    assert deviation < 1.e-06
    # clean up the output files in the test directory from this test.
    os.system('/bin/rm V88888024002P000000000112o*')



#! /usr/bin/env python
"""This is the module docstring.

The module docstring should have a one line description (as above) as
well as a more detailed description in a paragraph below the one line
description (i.e. this).  Module dosctring lines should be limited to
72 characters.  Monospace font can be achived with ``two single
forward-apostrophes``.  The description should provided a detailed
overview of the purpose of the module (what does it do) and how it
achieves this purpose (how does it do it).

Authors
-------

    - Bryan Hilbert

Use
---

    This module can be executed via the command line as such:

    ::
        python example.py [path] [-f|--filter filter]

    Required arguments:

    ``path`` - The path to the input file

    Optional arguments:

    ``-f|--filter`` - The filter to process.  if not provided, the
        defult value is "F606W".

Dependencies
------------

    Here is where any external dependencies can be listed or described.
    For example:

    The user must have a configuration file named ``config.yaml``
    placed in the current working directory.

References
----------

    Here is where any references to external sources related to the
    code can be listed or described.  For example:

    Code adopted from IDL routine written by Hilbert et al., 2009.

Notes
-----

    Here is where any additional notes (that are beyond the scope of the
    description) can be described.
"""

import argparse
import glob
import os
import sys
from typing import List, Union, Tuple, Optional, Any, Dict

from astropy.io import fits
import matplotlib.pyplot as plt
import numpy as np
import scipy
from sqlalchemy import Float, Integer, String


# Global variables should be avoided, but if used should be named with
# all-caps
A_GLOBAL_VARIABLE = 'foo' # type: str


def my_main_function(path: str, filter: str) -> None:
    """The main function of the ``example`` module.

    This function performs the main tasks of the module.  See module
    docstrings for further details.

    Parameters
    ----------
    path : str
        The path to the input file.
    filter : str
        The filter to process (e.g. "F606W").
    """

    print('Using {} as an input file'.format(path))
    an_int = 1 # type: int
    a_float = 3.14 # type: float
    a_bool = True # type: bool
    a_list = ['Dog', 'Cat', 'Turtle', False, 7] # type: List[Union[str, bool, int]]
    a_tuple = ('Dog', 'Cat', 'Turtle', False, 7) # type: Tuple[str, str, str, bool, int]
    a_dict = {'key1': 'value1', 'key2': 'value2'} # type: Dict[str, str]
    an_obj = object() # type: object

    result = some_other_function(an_int, a_float, a_bool, a_list, a_tuple, a_dict, an_obj) # type: Optional[int]

    print(result)


def parse_args() -> argparse.Namespace:
    """Parse command line arguments. Returns ``args`` object.

    Returns
    -------
    args : obj
        An argparse object containing all of the arguments
    """

    # Create help strings
    path_help = 'The path to the input file.' # type: str
    filter_help = 'The filter to process (e.g. "F606W").' # type: str

    # Add arguments
    parser = argparse.ArgumentParser() # type: argparse.ArgumentParser
    parser.add_argument('path',
                        type=str,
                        default=os.getcwd(),
                        help=path_help)
    parser.add_argument('-f --filter',
                        dest='filter',
                        type=str,
                        required=False,
                        default='F606W',
                        help=filter_help)

    # Parse args
    args = parser.parse_args() # type: argparse.Namespace

    return args


def some_other_function(an_int: int, a_float: float, a_bool: bool, a_list: List[Any],
                        a_tuple: Tuple[Any], a_dict: Dict[Any, Any], an_obj: object) -> int:
    """This function just does a bunch of nonsense.

    But it serves as a decent example of some things.

    Parameters
    ----------
    an_int : int
        Who knows what we will use this integer for.
    a_bool : bool
        Who knows what we will use this boolean for.
    a_float: float
        Who knows what we will use this float for.
    a_list : list
        Who knows what we will use this list for.
    a_tuple : tuple
        Who knows what we will use this tuple for.
    a_dict : dict
        Who knows what we will use this dictionary for.
    an_obj : obj
        Who knows what we will use this object for.

    Returns
    -------
    results : int
        The result of the function.
    """

    # File I/O should be handeled with 'with open' when possible
    with open('my_file', 'w') as f:
        f.write('My favorite integer is {}'.format(an_int))

    # Operators should be separated by spaces
    print(a_float + a_float)

    return an_int



if __name__ == '__main__':

    args = parse_args() # type: argparse.Namespace

    my_main_function(args.path, args.filter)


"""SIAF interface module to support accessing SIAF information.

This module provides ``mirage`` with functions to interface SIAF content via the pysiaf module.

Authors
-------

    - Johannes Sahlmann
    - Bryan Hilbert

Use
---

    This module can be imported and used with

    ::

        from mirage.utils import siaf_interface

"""
import os
import numpy as np

import pysiaf
from pysiaf import iando
from ..utils import rotations
from ..utils import set_telescope_pointing_separated as set_telescope_pointing


def aperture_ra_dec(siaf_instance, aperture_name, ra, dec, telescope_roll, output_apertures):
    """For a given aperture with a known RA, Dec, and telescope roll angle,
    calculate the RA, Dec values at the reference location in a list of
    other apertures.

    Parameters
    ----------
    siaf_instance : pysiaf.Siaf
        Instance of SIAF for a single instrument

    aperture_name : str
        Aperture name (e.g. "NRCA1_FULL")

    ra : float
        RA value of pointing in degrees

    dec : float
        Dec value of pointing in degrees

    telescope_roll : float
        PA_V3, Position angle of the telescope in degrees

    output_apertures : list
        List of aperture names to calculate RA, Dec for

    Returns
    -------
    aperture_pointing : dict
        Dictionary with output_apertures as keys. Values are (ra, dec)
        tuples
    """
    local_roll, att_matrix, fullframesize, subarray_boundaries = get_siaf_information(siaf_instance,
                                                                                      aperture_name,
                                                                                      ra, dec,
                                                                                      telescope_roll)
    aperture_pointing = {}
    for aperture in output_apertures:
        siaf_out = siaf_instance[aperture]
        out_ra, out_dec = pysiaf.rotations.pointing(att_matrix, siaf_out.V2Ref, siaf_out.V3Ref)
        aperture_pointing[aperture] = (out_ra, out_dec)
    return aperture_pointing


def get_instance(instrument):
    """Return an instance of a pysiaf.Siaf object for the given instrument

    Parameters
    ----------
    instrument : str
        Name of instrument

    Returns
    -------
    siaf : pysiaf.Siaf
        Siaf object for the requested instrument
    """
    siaf = pysiaf.Siaf(instrument)
    return siaf


def get_siaf_information(siaf_instance, aperture_name, ra, dec, telescope_roll, v2_arcsec=None,
                         v3_arcsec=None, verbose=False):
    """Use pysiaf to get aperture information.

    Parameters
    ----------
    siaf_instance : pysiaf.Siaf
        Instance of SIAF for a single instrument

    aperture_name : str
        Aperture name (e.g. "NRCA1_FULL")

    ra : float
        RA value of pointing in degrees

    dec : float
        Dec value of pointing in degrees

    telescope_roll : float
        PA_V3, Position angle of the telescope in degrees

    v2_arcsec : float
        The V2 value in arcseconds of the reference location for the
        instrument aperture

    v3_arcsecc : float
        The V3 value in arcseconds of the reference location for the
        instrument aperture

    verbose : bool
        Print extra information to the screen

    Returns
    -------
    local_roll : float
        Local roll angle at the reference location of the aperture

    att_matrix : matrix
        Attitude matrix used to relate RA, Dec, local roll angle to V2, V3

    fullframesize : int
        Number of columns in the  given aperture

    subarray_boundaries : list
        List of full-frame coordinates corresponding to the minimum and maximum
        values of x and y in the given aperture
    """
    # Select the correct aperture
    siaf = siaf_instance[aperture_name]

    if v2_arcsec is None:
        v2_arcsec = siaf.V2Ref
    if v3_arcsec is None:
        v3_arcsec = siaf.V3Ref

    local_roll = set_telescope_pointing.compute_local_roll(telescope_roll,
                                                           ra, dec, v2_arcsec, v3_arcsec)

    # Create attitude_matrix
    att_matrix = rotations.attitude(v2_arcsec, v3_arcsec, ra, dec, local_roll)

    # Get full frame size
    fullframesize = siaf.XDetSize

    # Subarray boundaries in full frame coordinates
    try:
        xcorner, ycorner = sci_subarray_corners(siaf_instance.instrument, aperture_name, siaf=siaf_instance)
        subarray_boundaries = [xcorner[0], ycorner[0], xcorner[1], ycorner[1]]
    except (RuntimeError, TypeError) as e:  # e.g. NIRSpec NRS_FULL_MSA aperture
        if verbose:
            print('get_siaf_information raised error:\n{}\nIgnoring it.'.format(e))
        subarray_boundaries = [0, 0, 0, 0]
    return local_roll, att_matrix, fullframesize, subarray_boundaries


def sci_subarray_corners(instrument, aperture_name, siaf=None, verbose=False):
    """Return the two opposing aperture corners in the SIAF Science frame of the full-frame SCA.

    This function serves as interface between the SIAF information accessible via the pysiaf package
    and the subarray information formatted for use by mirage.

    Parameters
    ----------
    instrument : str
        JWST instrument name with correct capitalization
    aperture_name : str
        SIAF aperture name
    siaf : pysiaf.Siaf
        SIAF instance for a single instrument
    verbose : bool
        Verbose output on/off

    Returns
    -------
    x_sci, y_sci : tuple of numpy arrays
        Subarray corner coordinates

    """
    # get SIAF
    if siaf is None:
        siaf = get_instance(instrument)

    # get master aperture names
    siaf_detector_layout = iando.read.read_siaf_detector_layout()
    master_aperture_names = siaf_detector_layout['AperName'].data

    # read pysiaf aperture definition file containing DetRef and SciRef values
    siaf_aperture_definitions = iando.read.read_siaf_aperture_definitions(instrument)

    # aperture object
    aperture = siaf[aperture_name]

    # aperture corners in SIAF detector coordinates
    x_det, y_det = aperture.corners('det', rederive=True)

    # determine parent aperture, i.e. the underlying full frame SCA aperture
    index = siaf_aperture_definitions['AperName'].tolist().index(aperture_name)
    aperture._parent_apertures = siaf_aperture_definitions['parent_apertures'][index]

    # If multiuple apertures are listed as parents keep only the first
    if ';' in aperture._parent_apertures:
        print('Multiple parent apertures: {}'.format(aperture._parent_apertures))
        aperture._parent_apertures = aperture._parent_apertures.split(';')[0]

    if aperture_name in master_aperture_names:
        # if master aperture, use it directly to transform to science frame
        x_sci, y_sci = aperture.det_to_sci(x_det, y_det)
    elif aperture._parent_apertures is not None:
        # use parent aperture for transformation
        if verbose:
            print('Using parent {} for {}'.format(aperture._parent_apertures, aperture_name))
        x_sci, y_sci = siaf[aperture._parent_apertures].det_to_sci(x_det, y_det)
        aperture = siaf[aperture._parent_apertures]

    if instrument.lower() == 'nircam':
        if aperture.DetSciParity == 1:
            corner_index = np.array([1, 3])
        elif aperture.DetSciParity == -1:
            # NIRCam will always fall in here, except in case of non-dms orientation
            corner_index = np.array([0, 2])
        x_corner = x_sci[corner_index]
        y_corner = y_sci[corner_index]
    elif instrument.lower() == 'niriss':
        x_corner_index = np.array([0, 2])
        y_corner_index = np.array([0, 2])
        if aperture_name == 'NIS_CEN_OSS':
            x_corner_index = np.array([1, 3])
            y_corner_index = np.array([3, 1])
        x_corner = x_sci[x_corner_index]
        y_corner = y_sci[y_corner_index]
        if aperture_name in ['NIS_SUBSTRIP96', 'NIS_SUBSTRIP256']:
            x_corner = [1, 2048]
            y_corner = [1, 2048]
    elif instrument.lower() == 'fgs':
        x_corner_index = np.array([0, 2])
        y_corner_index = np.array([0, 2])
        if aperture_name == 'FGS1_FULL_OSS':
            x_corner_index = np.array([1, 3])
            y_corner_index = np.array([3, 1])
        if aperture_name == 'FGS2_FULL_OSS':
            x_corner_index = np.array([1, 3])
            y_corner_index = np.array([1, 3])
        x_corner = x_sci[x_corner_index]
        y_corner = y_sci[y_corner_index]
    else:
        raise NotImplementedError(("Instrument {} not supported for SIAF subarray corners"
                                   .format(instrument)))

    # account for mirage conventions (e.g. 0-based indexing)
    # we also want integer values as these will be indexes
    x_corner = np.array([np.ceil(x_corner[0]) - 1, np.floor(x_corner[1]) - 1])
    y_corner = np.array([np.ceil(y_corner[0]) - 1, np.floor(y_corner[1]) - 1])
    return x_corner.astype(np.int), y_corner.astype(np.int)
